"""
===============
RandomCrop
===============
The :class:`~tonic.transforms.RandomCrop` crops the focal plane
to a smaller target size.
"""

import tonic

nmnist = tonic.datasets.NMNIST("../../tutorials/data", train=False)
events, label = nmnist[0]

transform = tonic.transforms.Compose(
    [
        tonic.transforms.RandomCrop(
            sensor_size=nmnist.sensor_size, target_size=(10, 10)
        ),
        tonic.transforms.ToFrame(
            sensor_size=(10, 10, 2),
            time_window=10000,
        ),
    ]
)

frames = transform(events)

ani = tonic.utils.plot_animation(frames)



import numpy as np


def decimate_numpy(events: np.ndarray, n: int):
    """Returns 1/n events for each pixel location.

    Parameters:
        events: structured numpy array of events
        n: filter rate
    """

    assert "x" in events.dtype.names
    assert n > 0, "n has to be an integer greater than zero."

    max_x = np.max(events["x"])

    output_events = []
    if "y" in events.dtype.names:
        max_y = np.max(events["y"])
        memory = np.zeros((max_x + 1, max_y + 1))

        for event in events:
            memory[event["x"], event["y"]] += 1
            if memory[event["x"], event["y"]] >= n:
                memory[event["x"], event["y"]] = 0
                output_events.append(event)

    else:
        memory = np.zeros((max_x + 1))

        for event in events:
            memory[event["x"]] += 1
            if memory[event["x"]] >= n:
                memory[event["x"]] = 0
                output_events.append(event)

    return np.array(output_events)



import os
from typing import Callable, Optional

import numpy as np

from tonic.dataset import Dataset


class POKERDVS(Dataset):
    """`POKER-DVS <http://www2.imse-cnm.csic.es/caviar/POKERDVS.html>`_

    Events have (txyp) ordering.
    ::

        @article{serrano2015poker,
          title={Poker-DVS and MNIST-DVS. Their history, how they were made, and other details},
          author={Serrano-Gotarredona, Teresa and Linares-Barranco, Bernab{\'e}},
          journal={Frontiers in neuroscience},
          volume={9},
          pages={481},
          year={2015},
          publisher={Frontiers}
        }

    Parameters:
        save_to (string): Location to save files to on disk.
        train (bool): If True, uses training subset, otherwise testing subset.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.
    """

    base_url = "https://nextcloud.lenzgregor.com/s/"
    train_filename = "pips_train.tar.gz"
    test_filename = "pips_test.tar.gz"
    train_url = base_url + "ZeCPYBS8kx4Wyjd/download/" + train_filename
    test_url = base_url + "2iRfwg3y9eAMpGL/download/" + test_filename
    train_md5 = "412bcfb96826e4fcb290558e8c150aae"
    test_md5 = "eef2bf7d0d3defae89a6fa98b07c17af"

    classes = ["cl", "he", "di", "sp"]
    int_classes = dict(zip(classes, range(4)))
    sensor_size = (35, 35, 2)
    dtype = np.dtype([("t", int), ("x", int), ("y", int), ("p", int)])
    ordering = dtype.names

    def __init__(
        self,
        save_to: str,
        train: bool = True,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        transforms: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )

        self.train = train

        if train:
            self.url = self.train_url
            self.file_md5 = self.train_md5
            self.filename = self.train_filename
            self.folder_name = "pips_train"
        else:
            self.url = self.test_url
            self.file_md5 = self.test_md5
            self.filename = self.test_filename
            self.folder_name = "pips_test"

        if not self._check_exists():
            self.download()

        file_path = os.path.join(self.location_on_system, self.folder_name)
        for path, dirs, files in os.walk(file_path):
            files.sort()
            for file in files:
                if file.endswith("npy"):
                    self.data.append(np.load(os.path.join(path, file)))
                    self.targets.append(self.int_classes[path[-2:]])

    def __getitem__(self, index):
        """
        Returns:
            a tuple of (events, target) where target is the index of the target class.
        """
        events, target = self.data[index], self.targets[index]
        events = np.lib.recfunctions.unstructured_to_structured(events, self.dtype)
        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        if self.transforms is not None:
            events, target = self.transforms(events, target)
        return events, target

    def __len__(self):
        return len(self.data)

    def _check_exists(self):
        return (
            self._is_file_present()
            and self._folder_contains_at_least_n_files_of_type(20, ".npy")
        )



import numpy as np
import pytest
from utils import create_random_input

import tonic.transforms as transforms


def testTimeReversalSpatialJitter():
    orig_events, sensor_size = create_random_input()

    flip_probability = 1
    var_x = 3
    var_y = 3
    sigma_xy = 0
    transform = transforms.Compose(
        [
            transforms.RandomTimeReversal(p=flip_probability),
            transforms.SpatialJitter(
                sensor_size=sensor_size,
                var_x=var_x,
                var_y=var_y,
                sigma_xy=sigma_xy,
                clip_outliers=False,
            ),
        ]
    )
    events = transform(orig_events)

    assert "RandomTimeReversal" in str(transform)
    assert "SpatialJitter" in str(transform)

    assert len(events) == len(orig_events), "Number of events should be the same."
    spatial_var_x = np.isclose(events["x"].all(), orig_events["x"].all(), atol=var_x)
    assert spatial_var_x, "Spatial jitter should be within chosen variance x."
    assert (events["x"] != orig_events["x"]).any(), "X coordinates should be different."
    spatial_var_y = np.isclose(events["y"].all(), orig_events["y"].all(), atol=var_y)
    assert spatial_var_y, "Spatial jitter should be within chosen variance y."
    assert (events["y"] != orig_events["y"]).any(), "Y coordinates should be different."
    assert np.array_equal(orig_events["p"], np.invert(events["p"][::-1].astype(bool)))
    assert np.array_equal(
        orig_events["t"], np.max(orig_events["t"]) - events["t"][::-1]
    )
    assert events is not orig_events


def testDropoutFlipUD():
    orig_events, sensor_size = create_random_input()

    flip_probability = 1
    drop_probability = 0.5

    transform = transforms.Compose(
        [
            transforms.DropEvent(p=drop_probability),
            transforms.RandomFlipUD(sensor_size=sensor_size, p=flip_probability),
        ]
    )

    events = transform(orig_events)

    drop_events = np.isclose(
        events.shape[0], (1 - drop_probability) * orig_events.shape[0]
    )
    assert drop_events, (
        "Event dropout should result in drop_probability*len(original) events"
        " dropped out."
    )

    temporal_order = np.isclose(np.sum((events["t"] - np.sort(events["t"])) ** 2), 0)
    assert temporal_order, "Temporal order should be maintained."

    first_dropped_index = np.where(events["t"][0] == orig_events["t"])[0][0]
    flipped_events = (
        sensor_size[1] - 1 - orig_events["y"][first_dropped_index] == events["y"][0]
    )
    assert flipped_events, (
        "When flipping up and down y must map to the opposite pixel, i.e. y' ="
        " sensor width - y"
    )
    assert events is not orig_events


def testTimeSkewFlipPolarityFlipLR():
    orig_events, sensor_size = create_random_input()

    coefficient = 1.5
    offset = 0
    flip_probability_pol = 1
    flip_probability_lr = 1

    transform = transforms.Compose(
        [
            transforms.TimeSkew(coefficient=coefficient, offset=offset),
            transforms.RandomFlipPolarity(p=flip_probability_pol),
            transforms.RandomFlipLR(sensor_size=sensor_size, p=flip_probability_lr),
        ]
    )

    events = transform(orig_events)

    assert len(events) == len(orig_events)
    assert (events["t"] >= orig_events["t"]).all()
    assert np.min(events["t"]) >= 0

    assert (
        events["p"] == np.invert(orig_events["p"].astype(bool)).astype(int)
    ).all(), "Polarities should be flipped."

    same_pixel = np.isclose((sensor_size[0] - 1) - events["x"][0], orig_events["x"][0])
    assert same_pixel, (
        "When flipping left and right x must map to the opposite pixel, i.e. x' ="
        " sensor width - x"
    )
    assert events is not orig_events



import numpy as np


def create_random_input(
    sensor_size=(200, 100, 2),
    n_events=10000,
    dtype=np.dtype([("x", int), ("y", int), ("t", int), ("p", int)]),
):
    """Creates random events for testing purposes.

    Returns
    - events - 10k events
    - sensor_size - 200 x 100 x 2 (w,h,pol)
    """

    assert "x" and "t" and "p" in dtype.names

    events = np.zeros(n_events, dtype=dtype)
    events["x"] = np.random.rand(n_events) * sensor_size[0]
    events["p"] = np.random.rand(n_events) * sensor_size[2]
    # sort timestamps to ensure the times are sequential
    events["t"] = np.sort(np.random.rand(n_events) * 1e6)

    if "y" in dtype.names:
        events["y"] = np.random.rand(n_events) * sensor_size[1]

    return events, sensor_size



import logging
import os
import shutil
from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable, Iterable, Optional, Tuple, Union
from warnings import warn

import h5py
import numpy as np


@dataclass
class MemoryCachedDataset:
    """MemoryCachedDataset caches the samples to memory to substantially improve data loading
    speeds. However you have to keep a close eye on memory consumption while loading your samples,
    which can increase rapidly when converting events to rasters/frames. If your transformed
    dataset doesn't fit into memory, yet you still want to cache samples to speed up training,
    consider using `DiskCachedDataset` instead.

    Parameters:
        dataset:
            Dataset to be cached to memory.
        device:
            Device to cache to. This is preferably a torch device. Will cache to CPU memory if None (default).
        transform:
            Transforms to be applied on the data
        target_transform:
            Transforms to be applied on the label/targets
        transforms:
            A callable of transforms that is applied to both data and labels at the same time.
    """

    dataset: Iterable
    device: Optional[str] = None
    transform: Optional[Callable] = None
    target_transform: Optional[Callable] = None
    transforms: Optional[Callable] = None
    samples_dict: dict = field(init=False, default_factory=dict)

    def __getitem__(self, index):
        try:
            data, targets = self.samples_dict[index]
        except KeyError as _:
            data, targets = self.dataset[index]
            if self.device is not None:
                data = data.to(self.device)
                targets = targets.to(self.device)
            self.samples_dict[index] = (data, targets)

        if self.transform is not None:
            data = self.transform(data)
        if self.target_transform is not None:
            targets = self.target_transform(targets)
        if self.transforms is not None:
            data, targets = self.transforms(data, targets)
        return data, targets

    def __len__(self):
        return len(self.dataset)


@dataclass
class DiskCachedDataset:
    """DiskCachedDataset caches the data samples to the hard drive for subsequent reads, thereby
    potentially improving data loading speeds. If dataset is None, then the length of this dataset
    will be inferred from the number of files in the caching folder. Pay attention to the cache
    path you're providing, as DiskCachedDataset will simply check if there is a file present with
    the index that it is looking for. When using train/test splits, it is wise to also take that
    into account in the cache path.

    .. note:: When you change the transform that is applied before caching, DiskCachedDataset cannot know about this and will present you
              with an old file. To avoid this you either have to clear your cache folder manually when needed, incorporate all
              transformation parameters into the cache path which creates a tree of cache files or use reset_cache=True.

    .. note:: Caching Pytorch tensors will write numpy arrays to disk, so be careful when loading the sample and you expect a tensor. The recommendation is to defer the transform to tensor as late as possible.

    Parameters:
        dataset:
            Dataset to be cached to disk. Can be None, if only files in cache_path should be used.
        cache_path:
            The preferred path where the cache will be written to and read from.
        reset_cache:
            When True, will clear out the cache path during initialisation. Default is False
        transform:
            Transforms to be applied on the data
        target_transform:
            Transforms to be applied on the label/targets
        transforms:
            A callable of transforms that is applied to both data and labels at the same time.
        num_copies:
            Number of copies of each sample to be cached.
            This is a useful parameter if the dataset is being augmented with slow, random transforms.
        compress:
            Whether to apply lightweight lzf compression, default is True.
    """

    dataset: Iterable
    cache_path: str
    reset_cache: bool = False
    transform: Optional[Callable] = None
    target_transform: Optional[Callable] = None
    transforms: Optional[Callable] = None
    num_copies: int = 1
    compress: bool = True

    def __post_init__(self):
        super().__init__()
        # Create cache directory
        if not os.path.isdir(self.cache_path):
            os.makedirs(self.cache_path)
        if self.reset_cache:
            shutil.rmtree(self.cache_path)
            os.makedirs(self.cache_path)
        if self.dataset is None:
            self.n_samples = (
                len(
                    [
                        name
                        for name in os.listdir(self.cache_path)
                        if os.path.isfile(os.path.join(self.cache_path, name))
                        and name.endswith(".hdf5")
                    ]
                )
                // self.num_copies
            )
        else:
            self.n_samples = len(self.dataset)

    def __getitem__(self, item) -> Tuple[object, object]:
        if self.dataset is None and item >= self.n_samples:
            raise IndexError(f"This dataset only has {self.n_samples} items.")

        copy = np.random.randint(self.num_copies)
        file_path = os.path.join(self.cache_path, f"{item}_{copy}.hdf5")
        try:
            data, targets = load_from_disk_cache(file_path)
        except (FileNotFoundError, OSError) as _:
            logging.info(
                f"Data {item}: {file_path} not in cache, generating it now",
                stacklevel=2,
            )

            data, targets = self.dataset[item]
            save_to_disk_cache(
                data, targets, file_path=file_path, compress=self.compress
            )
            # format might change during save to hdf5, i.e. tensors -> np arrays
            # We load the sample here again to keep the output format consistent.
            data, targets = load_from_disk_cache(file_path)

        if self.transform is not None:
            data = self.transform(data)
        if self.target_transform is not None:
            targets = self.target_transform(targets)
        if self.transforms is not None:
            data, targets = self.transforms(data, targets)
        return data, targets

    def __len__(self):
        return self.n_samples


def save_to_disk_cache(
    data, targets, file_path: Union[str, Path], compress: bool = True
) -> None:
    """
    Save data to caching path on disk in an hdf5 file. Can deal with data
    that is a dictionary.
    Args:
        data: numpy ndarray-like or a list thereof.
        targets: same as data, can be None.
        file_path: caching file path.
        compress: Whether to apply compression. (default = True - uses lzf compression)
    """
    with h5py.File(file_path, "w") as f:
        for name, data in zip(["data", "target"], [data, targets]):
            if type(data) != tuple:
                data = (data,)
            # can be events, frames, imu, gps, target etc.
            for i, data_piece in enumerate(data):
                if type(data_piece) == dict:
                    for key, item in data_piece.items():
                        f.create_dataset(
                            f"{name}/{i}/{key}",
                            data=item,
                            compression="lzf"
                            if type(item) == np.ndarray and compress
                            else None,
                        )
                else:
                    f.create_dataset(
                        f"{name}/{i}",
                        data=data_piece,
                        compression="lzf"
                        if type(data_piece) == np.ndarray and compress
                        else None,
                    )


def load_from_disk_cache(file_path: Union[str, Path]) -> Tuple:
    """Load data from file cache, separately for (data) and (targets).

    Can assemble dictionaries back together.
    Args:
        file_path: caching file path.
    Returns:
        data, targets
    """
    data_list = []
    target_list = []
    with h5py.File(file_path, "r") as f:
        for name, _list in zip(["data", "target"], [data_list, target_list]):
            for index in f[name].keys():
                if hasattr(f[f"{name}/{index}"], "keys"):
                    data = {
                        key: f[f"{name}/{index}/{key}"][()]
                        for key in f[f"{name}/{index}"].keys()
                    }
                else:
                    data = f[f"{name}/{index}"][()]
                _list.append(data)
    if len(data_list) == 1:
        data_list = data_list[0]
    if len(target_list) == 1:
        target_list = target_list[0]
    return data_list, target_list


class CachedDataset(DiskCachedDataset):
    """Deprecated class that points to DiskCachedDataset for now but will be removed in a future
    release.

    Please use MemoryCachedDataset or DiskCachedDataset in the future.
    """

    def __init__(self, *args, **kwargs):
        warn(
            "CachedDataset is deprecated and will be removed in a future release. "
            + "It currently points to DiskCachedDataset to distinguish it from "
            + "MemoryCachedDataset. Documentation available under https://tonic.readthedocs.io/en/latest/reference/data_classes.html#caching",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)



from .ncars import NCARS
from .nmnist import NMNIST
from .prophesee import (
    Gen1AutomotiveDetection,
    Gen4Automotive,
    Gen4AutomotiveDetectionMini,
    Gen4Downsampled,
)
from .stmnist import STMNIST

__all__ = [
    "NMNIST",
    "STMNIST",
    "NCARS",
    "Gen1AutomotiveDetection",
    "Gen4Automotive",
    "Gen4AutomotiveDetectionMini",
    "Gen4Downsampled",
]



import numpy as np


def crop_numpy(events, sensor_size, target_size):
    """Crops the sensor size to a smaller sensor.

    x' = x - new_sensor_start_x
    y' = y - new_sensor_start_y

    Parameters:
        events: ndarray of shape [num_events, num_event_channels]
        sensor_size: size of the sensor that was used [W,H]
        target_size: size of the sensor that was used [W',H']

    Returns:
        events - events within the crop box
        sensor_size - cropped to target_size
    """

    assert target_size[0] <= sensor_size[0] and target_size[1] <= sensor_size[1]
    assert "x" and "y" in events.dtype.names

    x_start_ind = int(np.random.rand() * (sensor_size[0] - target_size[0]))
    y_start_ind = int(np.random.rand() * (sensor_size[1] - target_size[1]))

    x_end_ind = x_start_ind + target_size[0]
    y_end_ind = y_start_ind + target_size[1]

    event_mask = (
        (events["x"] >= x_start_ind)
        * (events["x"] < x_end_ind)
        * (events["y"] >= y_start_ind)
        * (events["y"] < y_end_ind)
    )

    events = events[event_mask, ...]
    events["x"] -= x_start_ind
    events["y"] -= y_start_ind

    return events



import os
import warnings
from typing import Callable, List, Optional, Union

import h5py
import numpy as np

from tonic.dataset import Dataset
from tonic.download_utils import download_and_extract_archive, download_url, list_files
from tonic.io import make_structured_array


class DSEC(Dataset):
    """`DSEC <https://dsec.ifi.uzh.ch/>`_

    This is a fairly large dataset, so in order to save some disk space, event and image zips
    are deleted after extraction. If your download gets interrupted and you are left with a
    corrupted file on disk, Tonic will not be able to detect that and just proceed to download
    files that are not yet on disk. If you experience issues loading a particular recording,
    delete that folder manually and Tonic will re-download it the next time.
    Optical flow targets are not available for every recording, so if you select optical flow targets,
    only a subset of 18 training recordings will be selected.

    .. note:: To be able to read this dataset, you will need `hdf5plugin`, `PIL` and `imageio` packages installed.

    Parameters:
        save_to (str): Location to save files to on disk.
        split (str): Can be 'train', 'test' or a selection of individual recordings such as 'interlaken_00_c'
                     or ['thun_00_a', 'zurich_city_00_a']. Cannot mix across train/test.
        data_selection (str): Select which data to load per sample. Can be 'events_left', 'events_right',
                             'images_rectified_left', 'images_rectified_right', 'image_timestamps' or
                             any combination thereof in a list.
        target_selection (str, optional): Select which targets to load per sample. Can be 'disparity_events',
                                          'disparity_images', 'disparity_timestamps',
                                          'optical_flow_forward_event', 'optical_flow_forward_timestamps',
                                          'optical_flow_backward_event', 'optical_flow_backward_timestamps'
                                          or a combination thereof in a list. Note that optical flow targets
                                          are not available for every recording.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.
    """

    base_url = "https://download.ifi.uzh.ch/rpg/DSEC/"

    # boolean flag indicates optical flow availability
    recordings = {
        "train": {
            "interlaken_00_c": False,
            "interlaken_00_d": False,
            "interlaken_00_e": False,
            "interlaken_00_f": False,
            "interlaken_00_g": False,
            "thun_00_a": True,
            "zurich_city_00_a": False,
            "zurich_city_00_b": False,
            "zurich_city_01_a": True,
            "zurich_city_01_b": False,
            "zurich_city_01_c": False,
            "zurich_city_01_d": False,
            "zurich_city_01_e": False,
            "zurich_city_01_f": False,
            "zurich_city_02_a": True,
            "zurich_city_02_b": False,
            "zurich_city_02_c": True,
            "zurich_city_02_d": True,
            "zurich_city_02_e": True,
            "zurich_city_03_a": True,
            "zurich_city_04_a": False,
            "zurich_city_04_b": False,
            "zurich_city_04_c": False,
            "zurich_city_04_d": False,
            "zurich_city_04_e": False,
            "zurich_city_04_f": False,
            "zurich_city_05_a": True,
            "zurich_city_05_b": True,
            "zurich_city_06_a": True,
            "zurich_city_07_a": True,
            "zurich_city_08_a": True,
            "zurich_city_09_a": True,
            "zurich_city_09_b": False,
            "zurich_city_09_c": False,
            "zurich_city_09_d": False,
            "zurich_city_09_e": False,
            "zurich_city_10_a": True,
            "zurich_city_10_b": True,
            "zurich_city_11_a": True,
            "zurich_city_11_b": True,
            "zurich_city_11_c": True,
        },
        "test": [
            "thun_01_a",
            "thun_01_b",
            "interlaken_00_a",
            "interlaken_00_b",
            "interlaken_01_a",
            "zurich_city_12_a",
            "zurich_city_13_a",
            "zurich_city_13_b",
            "zurich_city_14_a",
            "zurich_city_14_b",
            "zurich_city_14_c",
            "zurich_city_15_a",
        ],
    }

    # that's a combination of the different data available, their
    # extension when downloaded and their extension when extracted
    data_names = {
        "events_left": [".zip", ".h5"],
        "events_right": [".zip", ".h5"],
        "image_timestamps": [".txt", ".txt"],
        "image_exposure_timestamps_left": [".txt", ".txt"],
        "image_exposure_timestamps_right": [".txt", ".txt"],
        "images_rectified_left": [".zip", ".png"],
        "images_rectified_right": [".zip", ".png"],
    }

    target_names = {
        "disparity_event": [".zip", ".png"],
        "disparity_image": [".zip", ".png"],
        "disparity_timestamps": [".txt", ".txt"],
        "optical_flow_forward_event": [".zip", ".png"],
        "optical_flow_forward_timestamps": [".txt", ".txt"],
        "optical_flow_backward_event": [".zip", ".png"],
        "optical_flow_backward_timestamps": [".txt", ".txt"],
    }

    sensor_size = (640, 480, 2)
    dtype = np.dtype([("x", np.int16), ("y", np.int16), ("t", np.int64), ("p", bool)])
    ordering = dtype.names

    def __init__(
        self,
        save_to: str,
        split: Union[str, List[str]],
        data_selection: Union[str, List[str]],
        target_selection: Optional[Union[str, List[str]]] = None,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        transforms: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )

        import imageio

        imageio.plugins.freeimage.download()

        if split == "train":
            self.recording_selection = self.recordings[split].keys()
            self.train_or_test = split

        elif split == "test":
            self.recording_selection = self.recordings[split]
            self.train_or_test = split

        else:
            if not isinstance(split, list):
                split = [split]

            for recording in split:
                if (
                    recording not in self.recordings["train"]
                    and recording not in self.recordings["test"]
                ):
                    raise RuntimeError(
                        f"Recording {recording} is neither in train nor in test set."
                    )
            self.recording_selection = split
            if all([recording in self.recordings["train"] for recording in split]):
                self.train_or_test = "train"
            elif all([recording in self.recordings["test"] for recording in split]):
                self.train_or_test = "test"
            else:
                raise RuntimeError("Cannot mix across train/test split.")
        self.train = self.train_or_test == "train"

        if isinstance(data_selection, str):
            data_selection = [data_selection]
        elif data_selection is None:
            data_selection = []
        self.data_selection = data_selection

        for data_piece in data_selection:
            if data_piece not in self.data_names.keys():
                raise RuntimeError(
                    f"Selection {data_piece} not available. Please select from the following options: {self.data_names.keys()}."
                )

        if isinstance(target_selection, str):
            target_selection = [target_selection]
        elif target_selection is None:
            target_selection = []
        self.target_selection = target_selection

        if not self.train and len(target_selection) > 0:
            raise Exception(
                "You wanted targets for the test set but they are not available."
            )

        for data_piece in target_selection:
            if data_piece not in self.target_names.keys():
                raise RuntimeError(
                    f"Selection {data_piece} not available. Please select from the following options: {self.target_names.keys()}."
                )

        # only take those recordings that have optical flow ground truth
        if any(["optical_flow" in selection for selection in target_selection]):
            deselect = [
                name
                for name in self.recording_selection
                if not self.recordings["train"][name]
            ]
            if len(deselect) > 0:
                warnings.warn(
                    f"Since you asked for optical flow targets, the following recordings without optical flow ground truth are dropped: {deselect}."
                )
                self.recording_selection = [
                    name
                    for name in self.recording_selection
                    if self.recordings["train"][name]
                ]

        self._check_exists(data_selection + target_selection)

    def __getitem__(self, index):
        """
        Returns:
            a tuple of (data, target) where data is another tuple of data_selction and target
            a tuple of target_selection if train=True.
        """
        import hdf5plugin  # necessary to read event files
        import imageio  # necessary to read optical flow pngs
        from PIL import Image  # necessary to read images

        recording = self.recording_selection[index]
        base_folder = os.path.join(self.location_on_system, recording)

        data_tuple = []
        for data_name in self.data_selection:
            full_base_folder = os.path.join(base_folder, data_name)
            if data_name in ["events_left", "events_right"]:
                with h5py.File(full_base_folder + "/events.h5", "r") as file:
                    data = make_structured_array(
                        file["events"]["x"][()],
                        file["events"]["y"][()],
                        file["events"]["t"][()],
                        file["events"]["p"][()],
                        dtype=self.dtype,
                    )
                    data["t"] += file["t_offset"][()]

            elif "images" in data_name:
                images_rectified_filenames = sorted(
                    list_files(full_base_folder, ".png", prefix=True)
                )
                data = np.stack(
                    [np.array(Image.open(file)) for file in images_rectified_filenames]
                )

            elif data_name == "image_timestamps":
                with open(full_base_folder + f"/{recording}_image_timestamps.txt") as f:
                    data = np.array([int(line) for line in f.readlines()])
            data_tuple.append(data)

        if self.transform is not None:
            data_tuple = self.transform(data_tuple)

        target_tuple = []
        for target_name in self.target_selection:
            full_base_folder = os.path.join(base_folder, target_name)
            if target_name in [
                "disparity_event",
                "disparity_image",
            ]:
                png_filenames = sorted(
                    list_files(full_base_folder, ".png", prefix=True)
                )
                target = np.stack(
                    [np.array(Image.open(file)) for file in png_filenames]
                )

            elif target_name in [
                "optical_flow_forward_event",
                "optical_flow_backward_event",
            ]:
                png_filenames = sorted(
                    list_files(full_base_folder, ".png", prefix=True)
                )
                target = np.array(
                    [imageio.v2.imread(file, format="PNG-FI") for file in png_filenames]
                ).astype(float)
                target[:, :, :, :2] -= 2**15
                target[:, :, :, :2] /= 128

            elif target_name == "disparity_timestamps":
                with open(full_base_folder + f"/{recording}_{target_name}.txt") as f:
                    target = np.array([int(line) for line in f.readlines()])

            elif target_name in [
                "optical_flow_forward_timestamps",
                "optical_flow_backward_timestamps",
            ]:
                with open(full_base_folder + f"/{recording}_{target_name}.txt") as f:
                    lines = f.readlines()
                    lines = lines[1:]  # first line is a comment
                    # first number is start timestamp, second number is stop timestamp
                    number_strs = [line.split(", ") for line in lines]
                    target = np.array(
                        [(int(start), int(stop)) for start, stop in number_strs]
                    )

            target_tuple.append(target)

        if self.target_transform is not None:
            target_tuple = self.target_transform(target_tuple)

        if self.transforms is not None:
            data_tuple, target_tuple = self.transforms(data_tuple, target_tuple)
        return data_tuple, target_tuple

    def __len__(self):
        return len(self.recording_selection)

    def _check_exists(self, data_selection: List):
        all_names = {**self.data_names, **self.target_names}
        for recording in self.recording_selection:
            for data_name in data_selection:
                file_folder = os.path.join(
                    self.location_on_system, recording, data_name
                )
                os.makedirs(file_folder, exist_ok=True)
                extension, extracted_file_extension = all_names[data_name]
                file_name = f"{recording}_{data_name + extension}"

                if any(
                    file.endswith(extracted_file_extension)
                    for file in os.listdir(file_folder)
                ):
                    continue

                url = f"{self.base_url}{self.train_or_test}/{recording}/{file_name}"
                if extension == ".zip":
                    download_and_extract_archive(url, file_folder)
                    os.remove(os.path.join(file_folder, file_name))
                else:
                    download_url(url, file_folder)



import numpy as np


def refractory_period_numpy(
    events: np.ndarray,
    refractory_period: float,
):
    """Sets a refractory period for each pixel, during which events will be ignored/discarded. We
    keep events if:

        .. math::
            t_n - t_{n-1} > t_{refrac}

    Parameters:
        events: ndarray of shape [num_events, num_event_channels]
        refractory_period: refractory period for each pixel in microseconds

    Returns:
        filtered set of events.
    """

    assert "t" and "x" and "y" in events.dtype.names

    events_copy = np.zeros_like(events)
    copy_index = 0
    width = int(events["x"].max()) + 1
    height = int(events["y"].max()) + 1
    timestamp_memory = np.zeros((width, height)) - refractory_period

    for event in events:
        time_since_last_spike = (
            event["t"] - timestamp_memory[int(event["x"]), int(event["y"])]
        )
        if time_since_last_spike > refractory_period:
            events_copy[copy_index] = event
            copy_index += 1
        timestamp_memory[int(event["x"]), int(event["y"])] = event["t"]

    return events_copy[:copy_index]



import os
from typing import Callable, Optional

import numpy as np

from tonic.dataset import Dataset
from tonic.io import read_mnist_file


class NCALTECH101(Dataset):
    """`N-CALTECH101 <https://www.garrickorchard.com/datasets/n-caltech101>`_

    Events have (xytp) ordering.
    ::

        @article{orchard2015converting,
          title={Converting static image datasets to spiking neuromorphic datasets using saccades},
          author={Orchard, Garrick and Jayawant, Ajinkya and Cohen, Gregory K and Thakor, Nitish},
          journal={Frontiers in neuroscience},
          volume={9},
          pages={437},
          year={2015},
          publisher={Frontiers}
        }

    Parameters:
        save_to (string): Location to save files to on disk.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.
    """

    url = "https://data.mendeley.com/public-files/datasets/cy6cvx3ryv/files/36b5c52a-b49d-4853-addb-a836a8883e49/file_downloaded"
    filename = "N-Caltech101-archive.zip"
    file_md5 = "66201824eabb0239c7ab992480b50ba3"
    data_filename = "Caltech101.zip"
    folder_name = "Caltech101"

    sensor_size = None  # all recordings are of different size
    dtype = np.dtype([("x", int), ("y", int), ("t", int), ("p", int)])
    ordering = dtype.names

    def __init__(
        self,
        save_to: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        transforms: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )

        if not self._check_exists():
            self.download()

        file_path = os.path.join(self.location_on_system, self.folder_name)
        for path, dirs, files in os.walk(file_path):
            dirs.sort()
            for file in files:
                if file.endswith("bin"):
                    self.data.append(path + "/" + file)
                    label_number = os.path.basename(path)
                    self.targets.append(label_number)

    def __getitem__(self, index):
        """
        Returns:
            a tuple of (events, target) where target is the index of the target class.
        """
        events = read_mnist_file(self.data[index], dtype=self.dtype)
        target = self.targets[index]
        events["x"] -= events["x"].min()
        events["y"] -= events["y"].min()
        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        if self.transforms is not None:
            events, target = self.transforms(events, target)
        return events, target

    def __len__(self):
        return len(self.data)

    def _check_exists(self):
        return (
            self._is_file_present()
            and self._folder_contains_at_least_n_files_of_type(8709, ".bin")
        )



import numpy as np
import pytest
from utils import create_random_input

import tonic.transforms as transforms


@pytest.mark.parametrize(
    "time_window, event_count, n_time_bins, n_event_bins, overlap,"
    " include_incomplete, sensor_size",
    [
        (20000, None, None, None, 0, False, (40, 20, 2)),
        (20000, None, None, None, 200, True, (40, 20, 1)),
        (10000, None, None, None, 100, True, (40, 20, 3)),
        (None, 2000, None, None, 0, False, (40, 20, 2)),
        (None, 2000, None, None, 200, True, (20, 20, 1)),
        (None, 2000, None, None, 100, True, (10, 20, 2)),
        (None, None, 5, None, 0, False, (40, 20, 2)),
        (None, None, 5, None, 0.1, False, (10, 20, 2)),
        (None, None, 5, None, 0.25, True, (40, 20, 2)),
        (None, None, None, 5, 0, True, (40, 20, 2)),
        (None, None, None, 5, 0.1, False, (40, 20, 2)),
        (None, None, None, 5, 0.25, False, (10, 20, 1)),
    ],
)
def test_representation_frame(
    time_window,
    event_count,
    n_time_bins,
    n_event_bins,
    overlap,
    include_incomplete,
    sensor_size,
):
    n_events = 10000
    orig_events, _ = create_random_input(sensor_size=sensor_size, n_events=n_events)

    transform = transforms.ToFrame(
        sensor_size=sensor_size,
        time_window=time_window,
        event_count=event_count,
        n_time_bins=n_time_bins,
        n_event_bins=n_event_bins,
        overlap=overlap,
        include_incomplete=include_incomplete,
    )

    frames = transform(orig_events)

    assert frames.shape[1:] == sensor_size[::-1]
    if time_window is not None:
        stride = time_window - overlap
        times = orig_events["t"]
        if include_incomplete:
            assert frames.shape[0] == int(
                np.ceil(((times[-1] - times[0]) - time_window) / stride) + 1
            )
        else:
            assert frames.shape[0] == int(
                np.floor(((times[-1] - times[0]) - time_window) / stride) + 1
            )

    if event_count is not None:
        assert event_count == frames[0].sum()
        stride = event_count - overlap
        if include_incomplete:
            assert frames.shape[0] == int(
                np.ceil((n_events - event_count) / stride) + 1
            )
        else:
            assert frames.shape[0] == int(
                np.floor((n_events - event_count) / stride) + 1
            )

    if n_time_bins is not None:
        assert frames.shape[0] == n_time_bins

    if n_event_bins is not None:
        assert frames.shape[0] == n_event_bins
        assert frames[0].sum() == (1 + overlap) * (n_events // n_event_bins)

    assert frames is not orig_events


@pytest.mark.parametrize(
    "time_window, event_count, n_time_bins, n_event_bins, overlap,"
    " include_incomplete, sensor_size",
    [
        (2000, None, None, None, 0, False, (40, 20, 2)),
        (2000, None, None, None, 200, True, (40, 20, 1)),
        (1000, None, None, None, 100, True, (40, 20, 3)),
        (None, 2000, None, None, 0, False, (40, 20, 2)),
        (None, 2000, None, None, 200, True, (20, 20, 1)),
        (None, 2000, None, None, 100, True, (10, 20, 2)),
        (None, None, 5, None, 0, False, (40, 20, 2)),
        (None, None, 5, None, 0.1, False, (10, 20, 2)),
        (None, None, 5, None, 0.25, True, (40, 20, 2)),
        (None, None, None, 5, 0, True, (40, 20, 2)),
        (None, None, None, 5, 0.1, False, (40, 20, 2)),
        (None, None, None, 5, 0.25, False, (10, 20, 1)),
    ],
)
def test_representation_sparse_tensor(
    time_window,
    event_count,
    n_time_bins,
    n_event_bins,
    overlap,
    include_incomplete,
    sensor_size,
):
    n_events = 10000
    orig_events, sensor_size = create_random_input(
        sensor_size=sensor_size, n_events=n_events
    )

    transform = transforms.ToSparseTensor(
        sensor_size=sensor_size,
        time_window=time_window,
        event_count=event_count,
        n_time_bins=n_time_bins,
        n_event_bins=n_event_bins,
        overlap=overlap,
        include_incomplete=include_incomplete,
    )

    sparse_tensor = transform(orig_events)

    assert sparse_tensor.is_sparse
    assert sparse_tensor.shape[1:] == sensor_size[::-1]

    if time_window is not None:
        stride = time_window - overlap
        times = orig_events["t"]
        if include_incomplete:
            assert sparse_tensor.shape[0] == int(
                np.ceil(((times[-1] - times[0]) - time_window) / stride) + 1
            )
        else:
            assert sparse_tensor.shape[0] == int(
                np.floor(((times[-1] - times[0]) - time_window) / stride) + 1
            )

    if event_count is not None:
        assert event_count == sparse_tensor[0].to_dense().sum()
        stride = event_count - overlap
        if include_incomplete:
            assert sparse_tensor.shape[0] == int(
                np.ceil((n_events - event_count) / stride) + 1
            )
        else:
            assert sparse_tensor.shape[0] == int(
                np.floor((n_events - event_count) / stride) + 1
            )

    if n_time_bins is not None:
        assert sparse_tensor.shape[0] == n_time_bins

    if n_event_bins is not None:
        assert sparse_tensor.shape[0] == n_event_bins
        assert sparse_tensor[0].to_dense().sum() == (1 + overlap) * (
            n_events // n_event_bins
        )

    assert sparse_tensor is not orig_events


def test_representation_inferred_frame():
    sensor_size = (20, 10, 2)
    orig_events, _ = create_random_input(n_events=30000, sensor_size=sensor_size)
    transform = transforms.ToFrame(sensor_size=None, time_window=25000)
    frames = transform(orig_events)
    assert frames.shape[1:] == sensor_size[::-1]


def test_representation_frame_wrong_sensor_size():
    sensor_size = (20, 10, 2)
    orig_events, _ = create_random_input(n_events=30000, sensor_size=sensor_size)
    transform = transforms.ToFrame(sensor_size=(20, 10, 1), time_window=25000)
    with pytest.raises(ValueError):
        frames = transform(orig_events)


def test_representation_audio_frame():
    sensor_size = (200, 1, 2)
    orig_events, _ = create_random_input(
        sensor_size=sensor_size,
        dtype=np.dtype([("x", int), ("t", int), ("p", int)]),
    )
    transform = transforms.ToFrame(sensor_size=sensor_size, time_window=25000)
    frames = transform(orig_events)
    assert frames.shape[1:] == (sensor_size[2], sensor_size[0])


def test_representation_image():
    sensor_size = (100, 100, 2)
    orig_events, _ = create_random_input(n_events=10000, sensor_size=sensor_size)
    transform = transforms.ToImage(sensor_size=sensor_size)
    image = transform(orig_events)
    assert image.shape == sensor_size[::-1]


@pytest.mark.parametrize(
    "sensor_size, dt, tau,",
    [
        ((40, 15, 2), 10000, 10000),
        ((10, 20, 2), 20000, 1000),
        ((30, 30, 2), 30000, 3000),
    ],
)
def test_representation_time_surface(sensor_size, dt, tau):
    orig_events, sensor_size = create_random_input(
        sensor_size=sensor_size, n_events=10000
    )

    transform = transforms.ToTimesurface(sensor_size=sensor_size, dt=dt, tau=tau)

    surfaces = transform(orig_events)

    duration = orig_events["t"][-1] - orig_events["t"][0]
    assert surfaces.shape[0] == duration // dt

    assert surfaces.shape[1] == 2
    assert surfaces.shape[2] == sensor_size[1]
    assert surfaces.shape[3] == sensor_size[0]
    assert surfaces is not orig_events


@pytest.mark.parametrize(
    "surface_size, cell_size, tau, decay",
    [(7, 9, 100, "lin"), (3, 4, 1000, "exp")],
)
def test_representation_avg_time_surface(surface_size, cell_size, tau, decay):
    orig_events, sensor_size = create_random_input(n_events=1000)

    transform = transforms.ToAveragedTimesurface(
        sensor_size=sensor_size,
        surface_size=surface_size,
        cell_size=cell_size,
        tau=tau,
        decay=decay,
    )

    surfaces = transform(orig_events)
    import math

    assert len(surfaces.shape) == 4
    assert surfaces.shape[0] == math.ceil(sensor_size[0] / cell_size) * math.ceil(
        sensor_size[1] / cell_size
    )
    assert surfaces.shape[1] == sensor_size[2]
    assert surfaces.shape[2] == surface_size
    assert surfaces.shape[3] == surface_size
    assert surfaces is not orig_events


@pytest.mark.parametrize("n_time_bins", [10, 1])
def test_representation_voxel_grid(n_time_bins):
    orig_events, sensor_size = create_random_input()

    transform = transforms.ToVoxelGrid(sensor_size=sensor_size, n_time_bins=n_time_bins)

    volume = transform(orig_events)

    assert volume.shape == (n_time_bins, 1, *sensor_size[1::-1])
    assert volume is not orig_events


@pytest.mark.parametrize(
    "n_frames, n_bits",
    [(1, 8), (2, 8), (3, 8), (1, 16), (2, 16)],
)
def test_bina_rep(n_frames, n_bits):
    n_events = 10000
    sensor_size = (128, 128, 2)

    orig_events, _ = create_random_input(sensor_size=sensor_size, n_events=n_events)

    transform = transforms.Compose(
        [
            transforms.ToFrame(sensor_size=sensor_size, n_time_bins=n_frames * n_bits),
            transforms.ToBinaRep(
                n_frames=n_frames,
                n_bits=n_bits,
            ),
        ]
    )

    frames = transform(orig_events)

    assert len(frames.shape) == 4
    assert frames.shape[0] == n_frames
    assert frames.shape[1:] == sensor_size[::-1]
    assert frames is not orig_events



import numpy as np


def to_bina_rep_numpy(
    event_frames: np.ndarray,
    n_frames: int = 1,
    n_bits: int = 8,
):
    """Representation that takes T*B binary event frames to produce a sequence of T frames of N-bit
    numbers. To do so, N binary frames are interpreted as a single frame of N-bit representation.
    Taken from the paper Barchid et al. 2022, Bina-Rep Event Frames: a Simple and Effective
    Representation for Event-based cameras https://arxiv.org/pdf/2202.13662.pdf.

    Parameters:
        event_frames: numpy.ndarray of shape (T*BxPxHxW). The sequence of event frames.
        n_frames (int): the number T of bina-rep frames.
        n_bits (int): the number N of bits used in the N-bit representation.

    Returns:
        (numpy.ndarray) the sequence of bina-rep event frames with dimensions (TxPxHxW).
    """
    assert type(event_frames) == np.ndarray and len(event_frames.shape) == 4
    assert n_frames >= 1
    assert n_bits >= 2

    if event_frames.shape[0] != n_bits * n_frames:
        raise ValueError(
            "the input event_frames must have the right number of frames to the targeted"
            f"sequence of {n_frames} bina-rep event frames of {n_bits}-bit representation."
            f"Got: {event_frames.shape[0]} frames. Expected: {n_frames}x{n_bits}={n_bits * n_frames} frames."
        )

    event_frames = (event_frames > 0).astype(np.float32)  # get binary event_frames

    bina_rep_seq = np.zeros((n_frames, *event_frames.shape[1:]), dtype=np.float32)

    for i in range(n_frames):
        frames = event_frames[i * n_bits : (i + 1) * n_bits]
        bina_rep_frame = bina_rep(frames)
        bina_rep_seq[i] = bina_rep_frame

    return bina_rep_seq


def bina_rep(frames: np.ndarray) -> np.ndarray:
    """Computes one Bina-Rep frame from the sequence of N binary event-frames in parameter.

    Args:
        frames (numpy.ndarray): the sequence of N binary event frames used to compute the bina-rep frame. Shape=(NxPxHxW)

    Returns:
        numpy.ndarray: the resulting bina-rep event frame. Shape=(PxHxW)
    """
    mask = 2 ** np.arange(frames.shape[0] - 1, -1, -1, dtype=np.float32)
    arr_mask = [
        mask for _ in range(frames.shape[1] * frames.shape[2] * frames.shape[3])
    ]
    mask = np.stack(arr_mask, axis=-1)
    mask = np.reshape(mask, frames.shape)

    return np.sum(mask * frames, 0) / (2 ** mask.shape[0] - 1)



"""
=======
ToImage
=======

:class:`~tonic.transforms.ToImage` bins all the events to a single image.
Typically used in conjunction with SlicedDataset, to chunk recordings
into smaller samples.

"""

import matplotlib.pyplot as plt

import tonic

nmnist = tonic.datasets.NMNIST("../../tutorials/data", train=False)
events, label = nmnist[0]

transform = tonic.transforms.ToImage(
    sensor_size=nmnist.sensor_size,
)

image = transform(events[:1000])

plt.imshow(image[1] - image[0])
plt.axis(False)
plt.show()



import shutil
from pathlib import Path

import numpy as np
import prototype_dataset_utils as dataset_utils

import tonic.prototype.datasets as datasets

PATH_TO_TEST_DATA = Path(".", "test", "test_data")

########
# NMNIST
########


def create_nmnist_zip_archive(tmpdir: str, folder: str, filename: str):
    """This function recreates the same structure that we expect in the dataset:

    [Train,Test]----!                 !---0/sample.bin                 !---1/sample.bin
    !---                 !---9/sample.bin
    """
    filename = filename.split(".")[0]  # Stripping the .zip extension.
    folder_path = Path(tmpdir, "NMNIST", folder)
    # We create a folder for each label.
    for i in range(10):
        destination_path = Path(folder_path, str(i))
        destination_path.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(
            PATH_TO_TEST_DATA / "sample_nmnist.bin",
            destination_path / "sample_nmnist.bin",
        )
    # We compress the archive.
    shutil.make_archive(
        base_name=Path(tmpdir, "NMNIST", filename),
        format="zip",
        root_dir=folder_path.parent,
        base_dir=folder,
    )
    # We remove the folder previously created.
    shutil.rmtree(folder_path)


class NMNISTTestCase_Train_Uncompressed_AllSaccades(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NMNIST
    FEATURE_TYPES = (datasets.NMNIST._DTYPE,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": True, "keep_compressed": False, "first_saccade_only": False}

    def inject_fake_data(self, tmpdir):
        create_nmnist_zip_archive(
            tmpdir, datasets.NMNIST._TRAIN_FOLDER, datasets.NMNIST._TRAIN_FILENAME
        )


class NMNISTTestCase_Train_Compressed_AllSaccades(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NMNIST
    FEATURE_TYPES = (datasets.NMNIST._DTYPE,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": True, "keep_compressed": True, "first_saccade_only": False}

    def inject_fake_data(self, tmpdir):
        create_nmnist_zip_archive(
            tmpdir, datasets.NMNIST._TRAIN_FOLDER, datasets.NMNIST._TRAIN_FILENAME
        )


class NMNISTTestCase_Train_Uncompressed_FirstSaccade(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NMNIST
    FEATURE_TYPES = (datasets.NMNIST._DTYPE,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": True, "keep_compressed": False, "first_saccade_only": True}

    def inject_fake_data(self, tmpdir):
        create_nmnist_zip_archive(
            tmpdir, datasets.NMNIST._TRAIN_FOLDER, datasets.NMNIST._TRAIN_FILENAME
        )


class NMNISTTestCase_Test_Uncompressed_AllSaccades(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NMNIST
    FEATURE_TYPES = (datasets.NMNIST._DTYPE,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": False, "keep_compressed": False, "first_saccade_only": False}

    def inject_fake_data(self, tmpdir):
        create_nmnist_zip_archive(
            tmpdir, datasets.NMNIST._TEST_FOLDER, datasets.NMNIST._TEST_FILENAME
        )


class NMNISTTestCase_Test_Compressed_AllSaccades(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NMNIST
    FEATURE_TYPES = (datasets.NMNIST._DTYPE,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": False, "keep_compressed": True, "first_saccade_only": False}

    def inject_fake_data(self, tmpdir):
        create_nmnist_zip_archive(
            tmpdir, datasets.NMNIST._TEST_FOLDER, datasets.NMNIST._TEST_FILENAME
        )


class NMNISTTestCase_Test_Uncompressed_FirstSaccade(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NMNIST
    FEATURE_TYPES = (datasets.NMNIST._DTYPE,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": False, "keep_compressed": False, "first_saccade_only": True}

    def inject_fake_data(self, tmpdir):
        create_nmnist_zip_archive(
            tmpdir, datasets.NMNIST._TEST_FOLDER, datasets.NMNIST._TEST_FILENAME
        )


##########
# STMNIST
##########


def create_stmnist_zip_archive(tmpdir: str, filename: str):
    """This function recreates the same structure that we expect in the dataset:
    data_submission/----!

    !---0/sample.mat !---1/sample.mat !--- !---9/sample.mat
    """
    folder_path = Path(tmpdir, "STMNIST", "data_submission")
    # We create a folder for each label.
    for i in range(10):
        destination_path = Path(folder_path, str(i))
        destination_path.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(
            PATH_TO_TEST_DATA / "sample_stmnist.mat",
            destination_path / "sample_stmnist.mat",
        )
    # We compress the archive.
    shutil.make_archive(
        base_name=Path(tmpdir, "STMNIST", filename),
        format="zip",
        base_dir=folder_path,
        root_dir=folder_path,
    )


class STMNISTTestCase_Uncompressed(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.STMNIST
    FEATURE_TYPES = (datasets.STMNIST._DTYPE,)
    TARGET_TYPES = (int,)
    KWARGS = {"keep_compressed": False}

    def inject_fake_data(self, tmpdir):
        create_stmnist_zip_archive(tmpdir, datasets.STMNIST.__name__)


class STMNISTTestCase_Compressed(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.STMNIST
    FEATURE_TYPES = (datasets.STMNIST._DTYPE,)
    TARGET_TYPES = (int,)
    KWARGS = {"keep_compressed": True}

    def inject_fake_data(self, tmpdir):
        create_stmnist_zip_archive(tmpdir, datasets.STMNIST.__name__)


#######
# NCARS
#######


def create_ncars_folder(tmpdir, folder_name):
    for label_folder_name in ("background", "cars"):
        test_dir = Path(tmpdir, "NCARS", folder_name, label_folder_name)
        test_dir.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(
            PATH_TO_TEST_DATA / "sample_ncars.dat",
            test_dir / "sample_ncars.dat",
        )


class NCARSTestCase_Train(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NCARS
    FEATURE_TYPES = (datasets.NCARS._DTYPE,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": True}

    def inject_fake_data(self, tmpdir):
        create_ncars_folder(tmpdir, datasets.NCARS._TRAIN_PATH)


class NCARSTestCase_Test(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NCARS
    FEATURE_TYPES = (datasets.NCARS._DTYPE,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": False}

    def inject_fake_data(self, tmpdir):
        create_ncars_folder(tmpdir, datasets.NCARS._TEST_PATH)


#############################
# Gen4AutomotiveDetectionMini
#############################


def create_minidataset_folder(tmpdir, split_folder):
    testfolder = Path(
        tmpdir, "Gen4AutomotiveDetectionMini", "mini_dataset", split_folder
    )
    testfolder.mkdir(parents=True, exist_ok=True)
    shutil.copyfile(
        PATH_TO_TEST_DATA / "sample_ncars.dat",
        Path(testfolder, "moorea_2019-02-19_004_td_610500000_670500000_td.dat"),
    )
    label = np.array([0.4, 0.2, 0.1, 0.1])
    np.save(testfolder / "moorea_2019-02-19_004_td_610500000_670500000_bbox.npy", label)


class Gen4AutomotiveDetectionMiniTestCase_Train(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.Gen4AutomotiveDetectionMini
    FEATURE_TYPES = (datasets.Gen4AutomotiveDetectionMini._DTYPE,)
    TARGET_TYPES = (np.array,)
    KWARGS = {"split": "train"}

    def inject_fake_data(self, tmpdir):
        create_minidataset_folder(
            tmpdir, datasets.Gen4AutomotiveDetectionMini._TRAIN_FOLDER
        )


class Gen4AutomotiveDetectionMiniTestCase_Valid(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.Gen4AutomotiveDetectionMini
    FEATURE_TYPES = (datasets.Gen4AutomotiveDetectionMini._DTYPE,)
    TARGET_TYPES = (np.array,)
    KWARGS = {"split": "valid"}

    def inject_fake_data(self, tmpdir):
        create_minidataset_folder(
            tmpdir, datasets.Gen4AutomotiveDetectionMini._VALID_FOLDER
        )


class Gen4AutomotiveDetectionMiniTestCase_Test(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.Gen4AutomotiveDetectionMini
    FEATURE_TYPES = (datasets.Gen4AutomotiveDetectionMini._DTYPE,)
    TARGET_TYPES = (np.array,)
    KWARGS = {"split": "test"}

    def inject_fake_data(self, tmpdir):
        create_minidataset_folder(
            tmpdir, datasets.Gen4AutomotiveDetectionMini._TEST_FOLDER
        )


#########################
# Gen1AutomotiveDetection
#########################


def create_gen1dataset_folder(tmpdir, split_folder):
    testfolder = Path(tmpdir, "detection_dataset_duration_60s_ratio_1.0", split_folder)
    testfolder.mkdir(parents=True, exist_ok=True)
    shutil.copyfile(
        PATH_TO_TEST_DATA / "sample_ncars.dat",
        Path(testfolder, "moorea_2019-02-19_004_td_610500000_670500000_td.dat"),
    )
    label = np.array([0.4, 0.2, 0.1, 0.1])
    np.save(testfolder / "moorea_2019-02-19_004_td_610500000_670500000_bbox.npy", label)


class Gen1AutomotiveDetectionTestCase_Train(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.Gen1AutomotiveDetection
    FEATURE_TYPES = (datasets.Gen1AutomotiveDetection._DTYPE,)
    TARGET_TYPES = (np.array,)
    KWARGS = {"split": "train"}

    def inject_fake_data(self, tmpdir):
        create_gen1dataset_folder(
            tmpdir, datasets.Gen1AutomotiveDetection._TRAIN_FOLDER
        )


class Gen1AutomotiveDetectionTestCase_Valid(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.Gen1AutomotiveDetection
    FEATURE_TYPES = (datasets.Gen1AutomotiveDetection._DTYPE,)
    TARGET_TYPES = (np.array,)
    KWARGS = {"split": "valid"}

    def inject_fake_data(self, tmpdir):
        create_gen1dataset_folder(
            tmpdir, datasets.Gen1AutomotiveDetection._VALID_FOLDER
        )


class Gen1AutomotiveDetectionTestCase_Test(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.Gen1AutomotiveDetection
    FEATURE_TYPES = (datasets.Gen1AutomotiveDetection._DTYPE,)
    TARGET_TYPES = (np.array,)
    KWARGS = {"split": "test"}

    def inject_fake_data(self, tmpdir):
        create_gen1dataset_folder(tmpdir, datasets.Gen1AutomotiveDetection._TEST_FOLDER)



from typing import Tuple

import numpy as np
from numpy.lib import recfunctions as rfn

from tonic.slicers import slice_events_by_time


def to_timesurface_numpy(
    events,
    sensor_size: Tuple[int, int, int],
    dt: float,
    tau: float,
    overlap: int = 0,
    include_incomplete: bool = False,
):
    """Representation that creates timesurfaces for each event in the recording. Modeled after the
    paper Lagorce et al. 2016, Hots: a hierarchy of event-based time-surfaces for pattern
    recognition https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=7508476. Unlike the paper,
    surfaces are always generated across the whole sensor, not just around the event.

    Parameters:
        sensor_size: x/y/p dimensions of the sensor
        dt: time interval at which the time-surfaces are accumulated
        tau (float): time constant to decay events around occuring event with.

    Returns:
        array of timesurfaces with dimensions (n_events//dt, p, h , w)
    """

    assert dt >= 0, print("Parameter delta_t cannot be negative.")

    event_slices = slice_events_by_time(
        events, time_window=dt, overlap=overlap, include_incomplete=include_incomplete
    )
    memory = np.zeros((sensor_size[::-1]), dtype=int)
    all_surfaces = []
    x_index = event_slices[0].dtype.names.index("x")
    y_index = event_slices[0].dtype.names.index("y")
    p_index = event_slices[0].dtype.names.index("p")
    t_index = event_slices[0].dtype.names.index("t")
    start_t = event_slices[0][0][t_index]
    for i, slice in enumerate(event_slices):
        # structured to unstructured in order to access the indices
        slice = rfn.structured_to_unstructured(slice, dtype=int)
        indices = slice[:, [p_index, y_index, x_index]].T
        timestamps = slice[:, t_index]
        memory[tuple(indices)] = timestamps
        diff = -((i + 1) * dt + start_t - memory)
        surf = np.exp(diff / tau)
        all_surfaces.append(surf)
    return np.array(all_surfaces)



import os
from typing import Callable, Optional

import numpy as np
from importRosbag.importRosbag import importRosbag

from tonic.dataset import Dataset
from tonic.download_utils import check_integrity, download_url
from tonic.io import make_structured_array


class MVSEC(Dataset):
    """`MVSEC <https://daniilidis-group.github.io/mvsec/>`_
    ::

        @article{zihao2018multi,
          title={The Multi Vehicle Stereo Event Camera Dataset: An Event Camera Dataset for 3D Perception},
          author={Zihao Zhu, Alex and Thakur, Dinesh and Ozaslan, Tolga and Pfrommer, Bernd and Kumar, Vijay and Daniilidis, Kostas},
          journal={arXiv e-prints},
          pages={arXiv--1801},
          year={2018}
        }

    Parameters:
        save_to (string): Location to save files to on disk.
        scene (string): Choose one of 4 scenes: outdoor_night, outdoor_day, indoor_flying, motorcycle.
                        If you already have the data on your system, make sure to place the .bag files in a subfolder
                        'MVSEC/{scene}/bag_files.bag'.
        transform (callable, optional): A callable of transforms to apply to events and / or images for both left and right cameras.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.
    """

    resources = {
        "outdoor_night": [
            ["outdoor_night1_data.bag", "534bea503649eeb2801316704d6ab041"],
            ["outdoor_night1_gt.bag", "7e169e4048307e01f1a7ba5931ca7d4d"],
            ["outdoor_night2_data.bag", "371ff73324ba94ecb368b4c220dc8e54"],
            ["outdoor_night2_gt.bag", "d235d28f1c93203a1d7738f8e7a67ca3"],
            ["outdoor_night3_data.bag", "fc40889a48de7b28e6e2506125b229ac"],
            ["outdoor_night3_gt.bag", "2bbdffc94f8dd54f71486fcecfc82fe6"],
        ],
        "motorcycle": [
            ["motorcycle_data.bag", "ae1d929563c63c4d15e0e3d3412d41c4"],
            ["motorcycle_gt.bag", "bebf1cd58837abd1ca625ba219b47388"],
        ],
        "indoor_flying": [
            ["indoor_flying1_data.bag", "fd01f35eb52a754e8195d478ed2a00a2"],
            ["indoor_flying1_gt.bag", "277b97ad46f9dba3896f651ca47297aa"],
            ["indoor_flying2_data.bag", "015ac75086b248167e9602b72485e0eb"],
            ["indoor_flying2_gt.bag", "3c7ba64cd7bede77b1809cc151e54ed2"],
            ["indoor_flying3_data.bag", "fda3408d3c72b2d7445540e5bdbd6396"],
            ["indoor_flying3_gt.bag", "cae5648c84ec09d0316a8d4805dee62e"],
            ["indoor_flying4_data.bag", "30ba3a744dcd4408fc4102e88c636acc"],
            ["indoor_flying4_gt.bag", "f050c886fb34f3890fcf85680e267a21"],
        ],
        "outdoor_day": [
            ["outdoor_day1_data.bag", "7438c34b71d08ff38f52cef68834e9be"],
            ["outdoor_day1_gt.bag", "36ec3dcd0a222c4c2102641a2dc91ff0"],
            ["outdoor_day2_data.bag", "536d20bc59720b995df49f925f96b74d"],
            ["outdoor_day2_gt.bag", "69fb399411d7098b3e2cf3850f593e7b"],
        ],
    }

    base_url = "http://visiondata.cis.upenn.edu/mvsec/"
    sensor_size = (346, 260, 2)
    dtype = np.dtype([("x", int), ("y", int), ("t", int), ("p", int)])
    ordering = dtype.names

    def __init__(
        self,
        save_to: str,
        scene: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        transforms: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )
        self.scene = scene
        if not scene in self.resources.keys():
            raise RuntimeError(
                f"Scene {scene} is not available or in the wrong format. Select one of: indoor_flying, outdoor_day, outdoor_night, motorcycle."
            )

        if not self._check_exists():
            self.download()

    def __getitem__(self, index):
        """
        Returns:
            tuple of (data, targets), where data is another tuple of (events_left, events_right, imu_left,
            imu_right, images_left, images_right) and targets is a tuple of (depth_rect_left,
            depth_rect_right, pose) for ground truths.
        """
        # decode data file
        filename = os.path.join(
            self.location_on_system,
            self.scene,
            self.resources[self.scene][index * 2][0],
        )
        topics = importRosbag(filename, log="ERROR")
        events_left = topics["/davis/left/events"]
        events_left["ts"] -= events_left["ts"][0]
        events_left["ts"] *= 1e6
        events_left = make_structured_array(
            events_left["x"],
            events_left["y"],
            events_left["ts"],
            events_left["pol"],
            dtype=self.dtype,
        )

        events_right = topics["/davis/right/events"]
        events_right["ts"] -= events_right["ts"][0]
        events_right["ts"] *= 1e6
        events_right = make_structured_array(
            events_right["x"],
            events_right["y"],
            events_right["ts"],
            events_right["pol"],
            dtype=self.dtype,
        )
        imu_left = topics["/davis/left/imu"]
        imu_right = topics["/davis/right/imu"]
        images_left = topics["/davis/left/image_raw"]
        images_left = np.stack(images_left["frames"])
        images_right = topics["/davis/right/image_raw"]
        images_right = np.stack(images_right["frames"])
        data = events_left, events_right, imu_left, imu_right, images_left, images_right

        # decode ground truth file
        filename = os.path.join(
            self.location_on_system,
            self.scene,
            self.resources[self.scene][index * 2 + 1][0],
        )
        topics = importRosbag(filename, log="ERROR")
        depth_left = topics["/davis/left/depth_image_raw"]
        depth_left = np.stack(depth_left["frames"])
        depth_right = topics["/davis/right/depth_image_raw"]
        depth_right = np.stack(depth_right["frames"])
        depth_rect_left = topics["/davis/left/depth_image_rect"]
        depth_rect_left = np.stack(depth_rect_left["frames"])
        depth_rect_right = topics["/davis/right/depth_image_rect"]
        depth_rect_right = np.stack(depth_rect_right["frames"])
        pose = topics["/davis/left/pose"]
        targets = depth_rect_left, depth_rect_right, pose

        if self.transform is not None:
            data = self.transform(data)
        if self.target_transform is not None:
            targets = self.transform(targets)
        if self.transforms is not None:
            data, targets = self.transforms(data, targets)
        return data, targets

    def __len__(self):
        # divided by two because of data and ground truth file per recording
        return len(self.resources[self.scene]) // 2

    def download(self):
        for (filename, md5_hash) in self.resources[self.scene]:
            download_url(
                url=os.path.join(self.base_url, self.scene, filename),
                root=os.path.join(self.location_on_system, self.scene),
                filename=filename,
                md5=md5_hash,
            )

    def _check_exists(self):
        files_present = list(
            [
                check_integrity(
                    os.path.join(self.location_on_system, self.scene, filename)
                )
                for (filename, md5_hash) in self.resources[self.scene]
            ]
        )
        return all(files_present)



import os
import zipfile
from pathlib import Path
from typing import Any, BinaryIO, Callable, Iterator, Optional, Tuple, Union

import numpy as np
from torchdata.datapipes.iter import (
    FileLister,
    FileOpener,
    Filter,
    IterDataPipe,
    Mapper,
    Saver,
    ZipArchiveLoader,
)

from tonic.download_utils import download_url
from tonic.io import read_mnist_file

from .utils._dataset import Dataset, Sample
from .utils._utils import check_sha256


class NMNISTFileReader(IterDataPipe[Sample]):
    def __init__(
        self,
        dp: Union[IterDataPipe[str], IterDataPipe[Tuple[str, BinaryIO]]],
        dtype: Optional[np.dtype] = np.dtype(
            [("x", int), ("y", int), ("t", int), ("p", int)]
        ),
        keep_compressed: Optional[bool] = False,
    ) -> None:
        self.dp = dp
        self.dtype = dtype
        self.keep_cmp = keep_compressed

    def __iter__(self) -> Iterator[Sample]:
        if self.keep_cmp:
            for fname, fdata in self.dp:
                yield (
                    read_mnist_file(fdata, self.dtype, is_stream=True),
                    self._get_target(fname),
                )
        else:
            for fname in self.dp:
                yield (
                    read_mnist_file(fname, self.dtype),
                    self._get_target(fname),
                )

    def _get_target(self, fname: str) -> int:
        return int(fname.split(os.sep)[-2])


class NMNIST(Dataset):
    """`N-MNIST <https://www.garrickorchard.com/datasets/n-mnist>`_

    Events have (xytp) ordering.
    ::

        @article{orchard2015converting,
          title={Converting static image datasets to spiking neuromorphic datasets using saccades},
          author={Orchard, Garrick and Jayawant, Ajinkya and Cohen, Gregory K and Thakor, Nitish},
          journal={Frontiers in neuroscience},
          volume={9},
          pages={437},
          year={2015},
          publisher={Frontiers}
        }

    Parameters:
        root (string): Location to save files to on disk.
        train (bool): If True, uses training subset, otherwise testing subset.
        first_saccade_only (bool): If True, only work with events of the first of three saccades.
                                   Results in about a third of the events overall.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.
    """

    _DTYPE = np.dtype([("x", int), ("y", int), ("t", int), ("p", int)])
    _BASE_URL = "https://data.mendeley.com/public-files/datasets/468j46mzdv/files/"
    _TRAIN_URL = _BASE_URL + "39c25547-014b-4137-a934-9d29fa53c7a0/file_downloaded"
    _TRAIN_FILENAME = "train.zip"
    _TRAIN_SHA256 = "1a54ee392a5e5082a0bef52911cd9211f63b950a4905ccd8890553804d3335f9"
    _TRAIN_FOLDER = "Train"
    _TEST_URL = _BASE_URL + "05a4d654-7e03-4c15-bdfa-9bb2bcbea494/file_downloaded"
    _TEST_FILENAME = "test.zip"
    _TEST_SHA256 = "6ecfd5d85dbb49a631961d8dc3101871c5be53e645004ee34064f6557d169f09"
    _TEST_FOLDER = "Test"
    sensor_size = (34, 34, 2)

    def __init__(
        self,
        root: os.PathLike,
        train: Optional[bool] = True,
        first_saccade_only: Optional[bool] = False,
        keep_compressed: Optional[bool] = False,
    ) -> None:
        self.train = train
        self.first_saccade_only = first_saccade_only
        self.filename = self._TRAIN_FILENAME if self.train else self._TEST_FILENAME
        self.subfolder = self._TRAIN_FOLDER if self.train else self._TEST_FOLDER
        sha256 = self._TRAIN_SHA256 if self.train else self._TEST_SHA256
        root = Path(root, self.__class__.__name__)
        if not self._check_exists(root / self.filename):
            self._download(root, self.filename, sha256)
        super().__init__(
            root=root,
            keep_compressed=keep_compressed,
        )

    def __len__(self) -> int:
        return 60_000 if self.train else 10_000

    def _filter(self, fname: str) -> bool:
        return fname.endswith(".bin")

    def _saccade_filter(self, events: np.ndarray):
        return events[events["t"] <= int(1e5)]

    def _download(self, root, filename, sha256) -> None:
        url = self._TRAIN_URL if self.train else self._TEST_URL
        download_url(url=url, root=root, filename=filename)
        check_sha256(root / filename, sha256)

    def _check_exists(self, folder) -> bool:
        return not folder.is_file()

    def _is_unzipped(self, folder) -> bool:
        if not folder.is_dir():
            return False
        dp = FileLister(str(folder), recursive=True).filter(self._filter)
        return len(list(dp)) >= 60_000 if self.train else 10_000

    def _datapipe(self) -> IterDataPipe[Sample]:
        if not self.keep_cmp:
            folder = self._root / self.subfolder
            if not self._is_unzipped(folder):
                with zipfile.ZipFile(self._root / self.filename, "r") as zip_file:
                    zip_file.extractall(self._root)
            dp = FileLister(str(folder), recursive=True)
            dp = dp.filter(self._filter)
        else:
            zip_filepath = self._root / self.filename
            dp = FileLister(str(zip_filepath))
            dp = FileOpener(dp, mode="rb")
            dp = ZipArchiveLoader(dp)
        # Reading data to structured NumPy array and integer target.
        dp = NMNISTFileReader(dp, keep_compressed=self.keep_cmp)
        # Filtering the first saccade.
        if self.first_saccade_only:
            dp = Mapper(dp, self._saccade_filter, input_col=0)
        return dp



import os
import shutil

from tonic import SlicedDataset
from tonic.datasets import POKERDVS
from tonic.slicers import SliceByEventCount


def test_sliced_dataset():
    dataset = POKERDVS(save_to="./data/", train=False)

    metadata_path = "./cache/metadata"
    if os.path.exists(metadata_path):
        shutil.rmtree(metadata_path)

    target_number = 0
    for data, label in dataset:
        target_number += len(data) // 200

    slicer = SliceByEventCount(event_count=200)
    sliced_dataset = SlicedDataset(dataset, slicer, metadata_path=metadata_path)

    for data, label in sliced_dataset:
        assert len(data) == 200
        assert type(label) == int

    assert len(sliced_dataset) == target_number



import numpy as np


def time_jitter_numpy(
    events: np.ndarray,
    std: float = 1,
    clip_negative: bool = False,
    sort_timestamps: bool = False,
):
    """Changes timestamp for each event by drawing samples from a Gaussian distribution and adding
    them to each timestamp.

    Parameters:
        events: ndarray of shape [num_events, num_event_channels]
        std: the standard deviation of the time jitter
        clip_negative: drops events that have negative timestamps
        sort_timestamps: sort the events by timestamps after jittering

    Returns:
        temporally jittered set of events.
    """

    assert "t" in events.dtype.names

    shifts = np.random.normal(0, std, len(events))

    events["t"] = events["t"] + shifts

    if clip_negative:
        events = np.delete(events, (np.where(events["t"] < 0)), axis=0)

    if sort_timestamps:
        events = events[np.argsort(events["t"])]

    return events



import os
from pathlib import Path
from typing import Iterator, Optional

import h5py
import numpy as np
from expelliarmus import Wizard
from expelliarmus.wizard.clib import event_t
from torchdata.datapipes.iter import FileLister, IterDataPipe, Shuffler

from tonic.download_utils import download_url

from .utils._dataset import Dataset, Sample
from .utils._utils import check_sha256


class AutomotiveDetectionFileReader(IterDataPipe[Sample]):
    def __init__(self, dp: IterDataPipe[str]) -> None:
        self.dp = dp
        self._wizard = Wizard(encoding="dat")

    def __iter__(self) -> Iterator[Sample]:
        for data_file_path, label_file_path in self.dp:
            yield (
                self._wizard.read(data_file_path),
                np.load(label_file_path),
            )


class AutomotiveDetectionBaseClass(Dataset):
    """Base class for Automotive Detection datasets."""

    _DTYPE = event_t
    _TRAIN_FOLDER = "train"
    _VALID_FOLDER = "val"
    _TEST_FOLDER = "test"

    def __init__(
        self,
        root: os.PathLike,
        split: str = "train",
        skip_sha256_check: Optional[bool] = True,
        shuffle=False,
    ) -> None:
        self.split = split
        self.do_shuffle = shuffle
        super().__init__(
            root=root,
            keep_compressed=False,
            skip_sha256_check=skip_sha256_check,
        )

    def _dat_filter(self, fname: str) -> bool:
        return fname.endswith(".dat")

    def _label_filter(self, fname: str) -> bool:
        return fname.endswith(".npy")

    def _check_exists(self) -> bool:
        base_path = Path(self._root, self._FOLDERNAME)
        split_mapping = {
            "train": self._TRAIN_FOLDER,
            "valid": self._VALID_FOLDER,
            "test": self._TEST_FOLDER,
        }
        split_folder = base_path / split_mapping[self.split]
        folder_exists = split_folder.is_dir()

        # Checking that some binary files are present.
        file_dp = FileLister(str(split_folder), recursive=True).filter(self._dat_filter)
        return True if folder_exists and len(list(file_dp)) > 0 else False

    def _datapipe(self) -> IterDataPipe[Sample]:
        split_folder = {
            "train": self._TRAIN_FOLDER,
            "valid": self._VALID_FOLDER,
            "test": self._TEST_FOLDER,
        }[self.split]
        fpath = Path(self._root, self._FOLDERNAME, split_folder)
        data_dp = FileLister(str(fpath), recursive=True).filter(self._dat_filter)
        label_dp = FileLister(str(fpath), recursive=True).filter(self._label_filter)

        dp = zip(data_dp, label_dp)
        if self.do_shuffle:
            dp = Shuffler(dp, buffer_size=1_000_000)
        return AutomotiveDetectionFileReader(dp)


class Gen1AutomotiveDetection(AutomotiveDetectionBaseClass):
    """`Gen1 Automotive Detection Dataset <https://www.prophesee.ai/2020/01/24/prophesee-
    gen1-automotive-detection-dataset/>`_

    To download the data, you'll need to agree to Prophesee's Terms and Conditions.

    Then, the steps to acquire the data can be as follows:

    Download the torrent file for the dataset
    ::
        wget https://dataset.prophesee.ai/index.php/s/uE0QGLaFAEQnPwy/download\?path\=%2F\&files\=ATIS%20Automotive%20Detection%20Dataset.torrent
        -O Gen1Prophesee.torrent

    Download the data using peer-to-peer connections. On Linux this can be done using `aria2c` on the command line
    ::
        aria2c Gen1Prophesee.torrent

    This will download several 7z archives for training and testing. We'll need to unpack them manually by looping over the 7z files and feeding them to 7z
    ::
        sudo apt-get install p7zip-full
        for i in *.7z; do 7z x $i; done

    Parameters:
        root (string): Location to decompressed archive.
        split (str): Can be 'train' (default), 'valid' or 'test'.
        shuffle (bool): If True, the dataset will be shuffled randomly.
    """

    _FOLDERNAME = "detection_dataset_duration_60s_ratio_1.0"
    sensor_size = dict(x=304, y=240, p=2)
    class_map = {
        0: "car",
        1: "pedestrian",
    }

    def __init__(self, root: os.PathLike, split: str = "train", shuffle: bool = False):
        super().__init__(root=root, split=split, shuffle=shuffle)
        if not self._check_exists():
            raise RuntimeError(
                "You need to download and extract the dataset manually. See the Tonic documentation for more details."
            )

    def __len__(self) -> int:
        return {
            "train": 1459,
            "valid": 429,
            "test": 470,
        }[self.split]


class Gen4AutomotiveDetectionMini(AutomotiveDetectionBaseClass):
    """`Gen4 Automotive Detection <https://www.prophesee.ai/2020/11/24/automotive-megapixel-event-
    based-dataset/>`_

    This datasets needs 'expelliarmus' installed on the system. Events have "txyp" ordering.
    ::

        @article{de2020large,
          title={A large scale event-based detection dataset for automotive},
          author={De Tournemire, Pierre and Nitti, Davide and Perot, Etienne and Migliore, Davide and Sironi, Amos},
          journal={arXiv preprint arXiv:2001.08499},
          year={2020}
        }

    .. note:: The hosting server is very flaky and often interrupts the download before it is completed. If you end up with anything smaller than 23GB on disk, delete and try again.

    Parameters:
        root (string): Location to decompressed archive.
        split (str): Can be 'train' (default), 'valid' or 'test'.
        shuffle (bool): If True, the dataset will be shuffled randomly.
    """

    _URL = "https://dataset.prophesee.ai/index.php/s/ScqMu02G5pdYKPh/download"
    _FILENAME = "mini_dataset.zip"
    _FOLDERNAME = "mini_dataset"
    _SHA256 = "a13fb1240c19f2e1dbf453cecbb9e0c3ac9a7a5ea3cfc5a4f88760fff4977449"

    sensor_size = dict(x=1280, y=720, p=2)
    class_map = {
        0: "pedestrian",
        1: "two wheeler",
        2: "car",
        3: "truck",
        4: "bus",
        5: "traffic sign",
        6: "traffic light",
    }

    def __init__(
        self,
        root: os.PathLike,
        split: str = "train",
        shuffle: bool = False,
        skip_sha256_check: Optional[bool] = True,
    ) -> None:
        super().__init__(
            root=Path(root, self.__class__.__name__),
            split=split,
            shuffle=shuffle,
            skip_sha256_check=skip_sha256_check,
        )
        if not self._check_exists():
            self._download()

    def __len__(self) -> int:
        return {
            "train": 4,
            "valid": 1,
            "test": 1,
        }[self.split]

    def _download(self):
        download_url(url=self._URL, root=self._root, filename=self._FILENAME)
        check_sha256(Path(self._root, self._FILENAME), self._SHA256)


class Gen4Automotive(AutomotiveDetectionBaseClass):
    """`Gen4 Automotive Detection <https://www.prophesee.ai/2020/11/24/automotive-megapixel-event-
    based-dataset/>`_

    This datasets needs 'expelliarmus' installed on the system. Events have "txyp" ordering.
    ::

        @article{de2020large,
          title={A large scale event-based detection dataset for automotive},
          author={De Tournemire, Pierre and Nitti, Davide and Perot, Etienne and Migliore, Davide and Sironi, Amos},
          journal={arXiv preprint arXiv:2001.08499},
          year={2020}
        }

    To download the data, you'll need to agree to Prophesee's Terms and Conditions.

    Then, the steps to acquire the data can be as follows:

    Download the torrent file for the dataset
    ::
        wget https://dataset.prophesee.ai/index.php/s/8HY0Bv4mOU4RzBm/download?path=%2F&files=Large_Automotive_Detection_Dataset.torrent
        -O Gen4Prophesee.torrent

    Download the data using peer-to-peer connections. On Linux this can be done using `aria2c` on the command line
    ::
        aria2c Gen4Prophesee.torrent

    This will download several 7z archives for training, validation and testing. We'll need to unpack them manually by looping over the 7z files and feeding them to 7z
    ::
        sudo apt-get install p7zip-full
        for i in *.7z; do 7z x $i; done

    Parameters:
        root (string): Location to decompressed archive.
        split (str): Can be 'train' (default), 'valid' or 'test'.
        shuffle (bool): If True, the dataset will be shuffled randomly.
    """

    _FOLDERNAME = "Large_Automotive_Detection_Dataset"
    sensor_size = dict(x=1280, y=720, p=2)
    class_map = {
        0: "pedestrian",
        1: "two wheeler",
        2: "car",
        3: "truck",
        4: "bus",
        5: "traffic sign",
        6: "traffic light",
    }

    def __init__(
        self,
        root: os.PathLike,
        split: str = "train",
        shuffle: bool = False,
    ) -> None:
        super().__init__(root=root, split=split, shuffle=shuffle)
        if not self._check_exists():
            raise RuntimeError(
                "You need to download and extract the dataset manually. See the Tonic documentation for more details."
            )

    def __len__(self) -> int:
        return {
            "train": 4,
            "valid": 1,
            "test": 1,
        }[self.split]


class Gen4Downsampled(Dataset):
    _FOLDERNAME = "Gen 4 Histograms"
    _TRAIN_FOLDER = "train"
    _VALID_FOLDER = "val"
    _TEST_FOLDER = "test"

    def __init__(
        self,
        root: os.PathLike,
        split: str = "train",
        skip_sha256_check: Optional[bool] = True,
        shuffle=False,
    ) -> None:
        self.split = split
        self.do_shuffle = shuffle
        super().__init__(
            root=root,
            keep_compressed=False,
            skip_sha256_check=skip_sha256_check,
        )

    def _dat_filter(self, fname: str) -> bool:
        return fname.endswith(".h5")

    def _label_filter(self, fname: str) -> bool:
        return fname.endswith(".npy")

    def _check_exists(self) -> bool:
        base_path = Path(self._root, self._FOLDERNAME)
        split_mapping = {
            "train": self._TRAIN_FOLDER,
            "valid": self._VALID_FOLDER,
            "test": self._TEST_FOLDER,
        }
        split_folder = base_path / split_mapping[self.split]
        folder_exists = split_folder.is_dir()

        # Checking that some binary files are present.
        file_dp = FileLister(str(split_folder), recursive=True).filter(self._dat_filter)
        return True if folder_exists and len(list(file_dp)) > 0 else False

    def _datapipe(self) -> IterDataPipe[Sample]:
        split_folder = {
            "train": self._TRAIN_FOLDER,
            "valid": self._VALID_FOLDER,
            "test": self._TEST_FOLDER,
        }[self.split]
        fpath = Path(self._root, self._FOLDERNAME, split_folder)
        data_dp = FileLister(str(fpath), recursive=True).filter(self._dat_filter)
        label_dp = FileLister(str(fpath), recursive=True).filter(self._label_filter)

        dp = zip(data_dp, label_dp)
        if self.do_shuffle:
            dp = Shuffler(dp, buffer_size=1_000_000)
        for data_file_path, label_file_path in dp:
            yield (
                h5py.File(data_file_path)["data"][()],
                np.load(label_file_path),
            )

    def __len__(self) -> int:
        return {
            "train": 705,
            "valid": 131,
            "test": 119,
        }[self.split]



"""
==========
DropPixel
==========
The :class:`~tonic.transforms.DropPixel` removes all events
that occur at given pixels.
"""

import numpy as np

import tonic

nmnist = tonic.datasets.NMNIST("../../tutorials/data", train=False)
events, label = nmnist[0]

transform = tonic.transforms.Compose(
    [
        tonic.transforms.DropPixel(
            coordinates=[
                [x, y]
                for x in np.random.randint(34, size=29)
                for y in np.random.randint(34, size=29)
            ]
        ),
        tonic.transforms.ToFrame(
            sensor_size=nmnist.sensor_size,
            time_window=10000,
        ),
    ]
)

frames = transform(events)

ani = tonic.utils.plot_animation(frames)



import os
import struct
from typing import BinaryIO, Optional, Union

import numpy as np
from numpy.lib import recfunctions

events_struct = np.dtype(
    [("x", np.int16), ("y", np.int16), ("t", np.int64), ("p", bool)]
)


# many functions in this file have been copied from https://gitlab.com/synsense/aermanager/-/blob/master/aermanager/parsers.py
def make_structured_array(*args, dtype=events_struct):
    """Make a structured array given a variable number of argument values.

    Parameters:
        *args: Values in the form of nested lists or tuples or numpy arrays.
               Every except the first argument can be of a primitive data type like int or float.

    Returns:
        struct_arr: numpy structured array with the shape of the first argument
    """
    assert not isinstance(
        args[-1], np.dtype
    ), "The `dtype` must be provided as a keyword argument."
    names = dtype.names
    assert len(args) == len(names)
    struct_arr = np.empty_like(args[0], dtype=dtype)
    for arg, name in zip(args, names):
        struct_arr[name] = arg
    return struct_arr


def read_aedat4(in_file):
    """Get the aer events from version 4 of .aedat file.

    Parameters:
        in_file: str The name of the .aedat file

    Returns:
        events:   numpy structured array of events
    """
    import aedat

    decoder = aedat.Decoder(in_file)
    target_id = None
    width = None
    height = None
    for stream_id, stream in decoder.id_to_stream().items():
        if stream["type"] == "events" and (target_id is None or stream_id < target_id):
            target_id = stream_id
    if target_id is None:
        raise Exception("there are no events in the AEDAT file")
    parsed_file = {
        "type": "dvs",
        "width": width,
        "height": height,
        "events": np.concatenate(
            tuple(
                packet["events"]
                for packet in decoder
                if packet["stream_id"] == target_id
            )
        ),
    }
    return parsed_file["events"]


def read_dvs_128(filename):
    """Get the aer events from DVS with resolution of rows and cols are (128, 128)

    Parameters:
        filename: filename

    Returns:
        shape (tuple):
            (height, width) of the sensor array
        xytp: numpy structured array of events
    """
    data_version, data_start = read_aedat_header_from_file(filename)
    all_events = get_aer_events_from_file(filename, data_version, data_start)
    all_addr = all_events["address"]
    t = all_events["timeStamp"]

    x = (all_addr >> 8) & 0x007F
    y = (all_addr >> 1) & 0x007F
    p = all_addr & 0x1

    xytp = make_structured_array(x, y, t, p)
    shape = (128, 128)
    return shape, xytp


def read_dvs_ibm(filename):
    """Get the aer events from DVS with ibm gesture dataset.

    Parameters:
        filename:   filename

    Returns:
        shape (tuple):
            (height, width) of the sensor array
        xytp: numpy structured array of events
    """
    data_version, data_start = read_aedat_header_from_file(filename)
    all_events = get_aer_events_from_file(filename, data_version, data_start)
    all_addr = all_events["address"]
    t = all_events["timeStamp"]

    x = (all_addr >> 17) & 0x00001FFF
    y = (all_addr >> 2) & 0x00001FFF
    p = (all_addr >> 1) & 0x00000001

    xytp = make_structured_array(x, y, t, p)
    shape = (128, 128)
    return shape, xytp


def read_dvs_red(filename):
    """Get the aer events from DVS with resolution of (260, 346)

    Parameters:
        filename:   filename

    Returns:
        shape (tuple):
            (height, width) of the sensor array

        events: numpy structured array of events
    """
    data_version, data_start = read_aedat_header_from_file(filename)
    all_events = get_aer_events_from_file(filename, data_version, data_start)
    all_addr = all_events["address"]
    t = all_events["timeStamp"]

    x = (all_addr >> 17) & 0x7FFF
    y = (all_addr >> 2) & 0x7FFF
    p = (all_addr >> 1) & 0x1

    xytp = make_structured_array(x, y, t, p)
    shape = (346, 260)
    return shape, xytp


def read_dvs_346mini(filename):
    """Get the aer events from DVS with resolution of (132,104)

    Parameters:
        filename: filename

    Returns:
        shape (tuple):
            (height, width) of the sensor array
        xytp: numpy structure of xytp
    """
    data_version, data_start = read_aedat_header_from_file(filename)
    all_events = get_aer_events_from_file(filename, data_version, data_start)
    all_addr = all_events["address"]
    t = all_events["timeStamp"]

    x = (all_addr >> 22) & 0x01FF
    y = (all_addr >> 12) & 0x03FF
    p = (all_addr >> 11) & 0x1

    xytp = make_structured_array(x, y, t, p)
    shape = (132, 104)
    return shape, xytp


def read_mnist_file(
    bin_file: Union[str, BinaryIO], dtype: np.dtype, is_stream: bool = False
):
    """Reads the events contained in N-MNIST/N-CALTECH101 datasets.

    Code adapted from https://github.com/gorchard/event-Python/blob/master/eventvision.py
    """
    if is_stream:
        raw_data = np.frombuffer(bin_file.read(), dtype=np.uint8).astype(np.uint32)
    else:
        with open(bin_file, "rb") as fp:
            raw_data = np.fromfile(fp, dtype=np.uint8).astype(np.uint32)

    all_y = raw_data[1::5]
    all_x = raw_data[0::5]
    all_p = (raw_data[2::5] & 128) >> 7  # bit 7
    all_ts = ((raw_data[2::5] & 127) << 16) | (raw_data[3::5] << 8) | (raw_data[4::5])

    # Process time stamp overflow events
    time_increment = 2**13
    overflow_indices = np.where(all_y == 240)[0]
    for overflow_index in overflow_indices:
        all_ts[overflow_index:] += time_increment

    # Everything else is a proper td spike
    td_indices = np.where(all_y != 240)[0]

    xytp = make_structured_array(
        all_x[td_indices],
        all_y[td_indices],
        all_ts[td_indices],
        all_p[td_indices],
        dtype=dtype,
    )
    return xytp


def read_aedat_header_from_file(filename):
    """Get the aedat file version and start index of the binary data.

    Parameters:
        filename (str):     The name of the .aedat file

    Returns:
        data_version (float):   The version of the .aedat file
        data_start (int):       The start index of the data
    """
    filename = os.path.expanduser(filename)
    assert os.path.isfile(filename), f"The .aedat file '{filename}' does not exist."
    f = open(filename, "rb")
    count = 1
    is_comment = "#" in str(f.read(count))

    while is_comment:
        # Read the rest of the line
        head = str(f.readline())
        if "!AER-DAT" in head:
            data_version = float(head[head.find("!AER-DAT") + 8 : -5])
        is_comment = "#" in str(f.read(1))
        count += 1
    data_start = f.seek(-1, 1)
    f.close()
    return data_version, data_start


def get_aer_events_from_file(filename, data_version, data_start):
    """Get aer events from an aer file.

    Parameters:
        filename (str):         The name of the .aedat file
        data_version (float):   The version of the .aedat file
        data_start (int):       The start index of the data

    Returns:
         all_events:          Numpy structured array:
                                  ['address'] the address of a neuron which fires
                                  ['timeStamp'] the timeStamp in mus when a neuron fires
    """
    filename = os.path.expanduser(filename)
    assert os.path.isfile(filename), "The .aedat file does not exist."
    f = open(filename, "rb")
    f.seek(data_start)

    if 2 <= data_version < 3:
        event_dtype = np.dtype([("address", ">u4"), ("timeStamp", ">u4")])
        all_events = np.fromfile(f, event_dtype)
    elif data_version > 3:
        event_dtype = np.dtype([("address", "<u4"), ("timeStamp", "<u4")])
        event_list = []
        while True:
            header = f.read(28)
            if not header or len(header) == 0:
                break

            # read header
            capacity = struct.unpack("I", header[16:20])[0]
            event_list.append(np.fromfile(f, event_dtype, capacity))
        all_events = np.concatenate(event_list)
    else:
        raise NotImplementedError()
    f.close()
    return all_events



from .asl_dvs import ASLDVS
from .cifar10dvs import CIFAR10DVS
from .davisdataset import DAVISDATA
from .dsec import DSEC
from .dvs_lips import DVSLip
from .dvsgesture import DVSGesture
from .hsd import SHD, SSC
from .mvsec import MVSEC
from .ncaltech101 import NCALTECH101
from .nmnist import NMNIST
from .pokerdvs import POKERDVS
from .s_mnist import SMNIST
from .tum_vie import TUMVIE
from .visual_place_recognition import VPR

__all__ = [
    "ASLDVS",
    "CIFAR10DVS",
    "DAVISDATA",
    "DSEC",
    "DVSGesture",
    "MVSEC",
    "NCALTECH101",
    "NMNIST",
    "POKERDVS",
    "SHD",
    "SMNIST",
    "SSC",
    "TUMVIE",
    "VPR",
    "DVSLip",
]



from dataclasses import dataclass
from typing import Any, List, Tuple

import numpy as np
from typing_extensions import Protocol, runtime_checkable


@runtime_checkable
class Slicer(Protocol):
    """Base protocol class for slicers in Tonic.

    That means that you don't have to directly inherit from it, but just implement its methods.
    """

    def get_slice_metadata(self, data: Any, targets: Any) -> List[Tuple[Any]]:
        """This method returns the metadata for each recording that helps with slicing, for example
        the indices or timestamps at which the data would be sliced. The return value is typically
        a list of tuples that contain start and stop information for each slice.

        Parameters:
            data: Normally a tuple of data pieces.
            target: Normally a tuple of target pieces.

        Returns:
            metadata as a list of tuples of start and end indices, timestamps, etc.
        """
        ...

    def slice_with_metadata(self, data: Any, targets: Any, metadata: Any) -> List[Any]:
        """Given a piece of data and/or targets, cut out a certain part of it based on the
        start/end information given in metadata.

        Parameters:
            data: Normally a tuple of data pieces.
            target: Normally a tuple of target pieces.
            metadata: An array that contains start and stop information about one slice.

        Returns:
            A subset of the original data/targets which is a slice.
        """
        ...

    def slice(self, data: Any, targets: Any) -> List[Any]:
        """Generate metadata and return all slices at once.

        Parameters:
            data: Normally a tuple of data pieces.
            target: Normally a tuple of target pieces.

        Returns:
            The whole data and targets sliced into smaller slices.
        """
        ...


@dataclass(frozen=True)
class SliceByTime:
    """Slices an event array along fixed time window and overlap size. The number of bins depends
    on the length of the recording. Targets are copied.

    >        <overlap>
    >|    window1     |
    >        |   window2     |

    Parameters:
        time_window (int): time for window length (same unit as event timestamps)
        overlap (int): overlap (same unit as event timestamps)
        include_incomplete (bool): include the last incomplete slice that has shorter time
    """

    time_window: float
    overlap: float = 0.0
    include_incomplete: bool = False

    def slice(self, data: np.ndarray, targets: int) -> List[np.ndarray]:
        metadata = self.get_slice_metadata(data, targets)
        return self.slice_with_metadata(data, targets, metadata)

    def get_slice_metadata(
        self, data: np.ndarray, targets: int
    ) -> List[Tuple[int, int]]:
        t = data["t"]
        stride = self.time_window - self.overlap
        assert stride > 0

        if self.include_incomplete:
            n_slices = int(np.ceil(((t[-1] - t[0]) - self.time_window) / stride) + 1)
        else:
            n_slices = int(np.floor(((t[-1] - t[0]) - self.time_window) / stride) + 1)
        n_slices = max(n_slices, 1)  # for strides larger than recording time

        window_start_times = np.arange(n_slices) * stride + t[0]
        window_end_times = window_start_times + self.time_window
        indices_start = np.searchsorted(t, window_start_times)[:n_slices]
        indices_end = np.searchsorted(t, window_end_times)[:n_slices]
        return list(zip(indices_start, indices_end))

    @staticmethod
    def slice_with_metadata(
        data: np.ndarray, targets: int, metadata: List[Tuple[int, int]]
    ):
        return [data[start:end] for start, end in metadata], targets


@dataclass(frozen=True)
class SliceByTimeBins:
    """
    Slices data and targets along fixed number of bins of time length time_duration / bin_count * (1 + overlap).
    This method is good if your recordings all have roughly the same time length and you want an equal
    number of bins for each recording. Targets are copied.

    Parameters:
        bin_count (int): number of bins
        overlap (float): overlap specified as a proportion of a bin, needs to be smaller than 1. An overlap of 0.1
                    signifies that the bin will be enlarged by 10%. Amount of bins stays the same.
    """

    bin_count: int
    overlap: float = 0

    def slice(self, data: np.ndarray, targets: int) -> List[np.ndarray]:
        metadata = self.get_slice_metadata(data, targets)
        return self.slice_with_metadata(data, targets, metadata)

    def get_slice_metadata(
        self, data: np.ndarray, targets: int
    ) -> List[Tuple[int, int]]:
        events = data
        assert "t" in events.dtype.names
        assert self.overlap < 1

        times = events["t"]
        time_window = (times[-1] - times[0]) // self.bin_count * (1 + self.overlap)
        stride = time_window * (1 - self.overlap)

        window_start_times = np.arange(self.bin_count) * stride + times[0]
        window_end_times = window_start_times + time_window
        indices_start = np.searchsorted(times, window_start_times)
        indices_end = np.searchsorted(times, window_end_times)
        return list(zip(indices_start, indices_end))

    @staticmethod
    def slice_with_metadata(
        data: np.ndarray, targets: int, metadata: List[Tuple[int, int]]
    ):
        return [data[start:end] for start, end in metadata], targets


@dataclass(frozen=True)
class SliceByEventCount:
    """Slices data and targets along a fixed number of events and overlap size. The number of bins
    depends on the amount of events in the recording. Targets are copied.

    Parameters:
        event_count (int): number of events for each bin
        overlap (int): overlap in number of events
        include_incomplete (bool): include the last incomplete slice that has fewer events
    """

    event_count: int
    overlap: int = 0
    include_incomplete: bool = False

    def slice(self, data: np.ndarray, targets: int) -> List[np.ndarray]:
        metadata = self.get_slice_metadata(data, targets)
        return self.slice_with_metadata(data, targets, metadata)

    def get_slice_metadata(
        self, data: np.ndarray, targets: int
    ) -> List[Tuple[int, int]]:
        n_events = len(data)
        event_count = min(self.event_count, n_events)

        stride = self.event_count - self.overlap
        if stride <= 0:
            raise Exception("Inferred stride <= 0")

        if self.include_incomplete:
            n_slices = int(np.ceil((n_events - event_count) / stride) + 1)
        else:
            n_slices = int(np.floor((n_events - event_count) / stride) + 1)

        indices_start = (np.arange(n_slices) * stride).astype(int)
        indices_end = indices_start + event_count
        return list(zip(indices_start, indices_end))

    @staticmethod
    def slice_with_metadata(
        data: np.ndarray, targets: int, metadata: List[Tuple[int, int]]
    ):
        return [data[start:end] for start, end in metadata], targets


@dataclass(frozen=True)
class SliceByEventBins:
    """
    Slices an event array along fixed number of bins that each have n_events // bin_count * (1 + overlap) events.
    This slicing method is good if you recordings have all roughly the same amount of overall activity in the scene
    and you want an equal number of bins for each recording. Targets are copied.

    Parameters:
        bin_count (int): number of bins
        overlap (float): overlap in proportion of a bin, needs to be smaller than 1. An overlap of 0.1
                    signifies that the bin will be enlarged by 10%. Amount of bins stays the same.
    """

    bin_count: int
    overlap: float = 0

    def slice(self, data: np.ndarray, targets: int) -> List[np.ndarray]:
        metadata = self.get_slice_metadata(data, targets)
        return self.slice_with_metadata(data, targets, metadata)

    def get_slice_metadata(
        self, data: np.ndarray, targets: int
    ) -> List[Tuple[int, int]]:
        n_events = len(data)
        spike_count = int(n_events // self.bin_count * (1 + self.overlap))
        stride = int(spike_count * (1 - self.overlap))

        indices_start = np.arange(self.bin_count) * stride
        indices_end = indices_start + spike_count
        return list(zip(indices_start, indices_end))

    @staticmethod
    def slice_with_metadata(
        data: np.ndarray, targets: int, metadata: List[Tuple[int, int]]
    ):
        return [data[start:end] for start, end in metadata], targets


@dataclass
class SliceAtIndices:
    """Slices data at the specified event indices. Targets are copied.

    Parameters:
        start_indices (list): List of start indices
        end_indices (list): List of end indices (exclusive)
    """

    start_indices: np.ndarray
    end_indices: np.ndarray

    def slice(self, data: np.ndarray, targets: int) -> List[np.ndarray]:
        metadata = self.get_slice_metadata(data, targets)
        return self.slice_with_metadata(data, targets, metadata)

    def get_slice_metadata(
        self, data: np.ndarray, targets: int
    ) -> List[Tuple[int, int]]:
        return list(zip(self.start_indices, self.end_indices))

    @staticmethod
    def slice_with_metadata(
        data: np.ndarray, targets: int, metadata: List[Tuple[int, int]]
    ):
        return [data[start:end] for start, end in metadata], targets


@dataclass
class SliceAtTimePoints:
    """Slice the data at the specified time points.

    Parameters:
        tw_start (list): List of start times
        tw_end (list): List of end times
    """

    start_tw: np.ndarray
    end_tw: np.ndarray

    def slice(self, data: np.ndarray, targets: int) -> List[np.ndarray]:
        metadata = self.get_slice_metadata(data, targets)
        return self.slice_with_metadata(data, targets, metadata)

    def get_slice_metadata(
        self, data: np.ndarray, targets: int
    ) -> List[Tuple[int, int]]:
        t = data["t"]
        indices_start = np.searchsorted(t, self.start_tw)
        indices_end = np.searchsorted(t, self.end_tw)
        return list(zip(indices_start, indices_end))

    @staticmethod
    def slice_with_metadata(
        data: np.ndarray, targets: int, metadata: List[Tuple[int, int]]
    ):
        return [data[start:end] for start, end in metadata], targets


def slice_events_by_time(
    events: np.ndarray,
    time_window: int,
    overlap: int = 0,
    include_incomplete: bool = False,
):
    return SliceByTime(
        time_window=time_window, overlap=overlap, include_incomplete=include_incomplete
    ).slice(events, None)[0]


def slice_events_by_time_bins(events: np.ndarray, bin_count: int, overlap: float = 0.0):
    return SliceByTimeBins(bin_count=bin_count, overlap=overlap).slice(events, None)[0]


def slice_events_by_count(
    events: np.ndarray,
    event_count: int,
    overlap: int = 0,
    include_incomplete: bool = False,
):
    return SliceByEventCount(
        event_count=event_count, overlap=overlap, include_incomplete=include_incomplete
    ).slice(events, None)[0]


def slice_events_by_event_bins(
    events: np.ndarray, bin_count: int, overlap: float = 0.0
):
    return SliceByEventBins(bin_count=bin_count, overlap=overlap).slice(events, None)[0]


def slice_events_at_indices(events: np.ndarray, start_indices, end_indices):
    return SliceAtIndices(start_indices=start_indices, end_indices=end_indices).slice(
        events, None
    )[0]


def slice_events_at_timepoints(
    events: np.ndarray, start_tw: np.ndarray, end_tw: np.ndarray
) -> List[np.ndarray]:
    return SliceAtTimePoints(start_tw=start_tw, end_tw=end_tw).slice(events, None)[0]



import tonic


def test_read_aedat_header():
    data_version, data_start = tonic.io.read_aedat_header_from_file(
        "test/test_data/sample.aedat4"
    )

    assert data_version == 4
    assert data_start == 14


def test_read_aedat_events():
    data_version, data_start = tonic.io.read_aedat_header_from_file(
        "test/test_data/sample.aedat4"
    )
    events = tonic.io.get_aer_events_from_file(
        "test/test_data/sample.aedat4", data_version, data_start
    )
    assert len(events) == 118651


def test_read_aedat4():
    events = tonic.io.read_aedat4("test/test_data/sample.aedat4")



#! /usr/bin/env python3

from math import ceil

import numpy as np


def _get_ts(event, locmem, time_window, tau, surface_size, decay):
    rho = surface_size // 2
    t_i, t_j = event["t"].astype(np.float32), locmem["t"].astype(np.float32)
    # Starting time stamp, calculated subtracting the time window from the event timestamp.
    t_start = max(0, t_i - time_window)
    # Relative coordinates in the time surfaces.
    ts_x, ts_y = locmem["x"] - event["x"], locmem["y"] - event["y"]
    # Including only the events in the neighbourhood and in the time window.
    mask = np.asarray(
        (np.abs(ts_x) <= rho) & (np.abs(ts_y) <= rho) & (t_j >= t_start)
    ).nonzero()[0]
    # For each event in the local memory that belongs to the spatial and temporal windows and for the current event, a time surface is generated.
    locmem_ts = np.zeros((1 + len(mask), surface_size, surface_size), dtype=np.float32)
    if len(mask) > 0:
        locmem_ts[np.arange(len(mask)), ts_y[mask] + rho, ts_x[mask] + rho] = (
            np.exp(-(t_i - t_j[mask]) / tau)
            if decay == "exp"
            else -(t_i - t_j[mask]) / (3 * tau) + 1
        )
    # Adding the current event time surface.
    locmem_ts[-1, rho, rho] += 1
    # The accumulated time surfaces are returned.
    return np.sum(locmem_ts, axis=0)


def _map_to_locmems(events, sensor_size, cell_size):
    w, h, npols = sensor_size
    wgrid, hgrid = ceil(w / cell_size), ceil(h / cell_size)
    ncells = int(wgrid * hgrid)
    px_to_cell = lambda y, x: int((y // cell_size) * wgrid + x // cell_size)
    # Mapping events to local memories.
    locmems = [[[] for p in range(npols)] for c in range(ncells)]
    for event in events:
        locmems[px_to_cell(int(event["y"]), int(event["x"]))][
            max(0, int(event["p"]))
        ].append(event)
    # Converting the lists in structured NumPy arrays.
    locmems = [
        [
            np.stack(locmems[c][p]) if locmems[c][p] else np.empty((0,))
            for p in range(npols)
        ]
        for c in range(ncells)
    ]
    return locmems


def to_averaged_timesurface_numpy(
    events,
    sensor_size,
    cell_size,
    surface_size,
    time_window,
    tau,
    decay,
):
    """Representation that creates averaged timesurfaces for each event for one recording.

    Taken from the paper
    Sironi et al. 2018, HATS: Histograms of averaged time surfaces for robust event-based object classification
    https://openaccess.thecvf.com/content_cvpr_2018/papers/Sironi_HATS_Histograms_of_CVPR_2018_paper.pdf
    Parameters:
        cell_size (int): size of each square in the grid
        surface_size (int): has to be odd
        time_window (int): how far back to look for past events for the time averaging. Expressed in microseconds.
        tau (int): time constant to decay events around occuring event with. Expressed in microseconds.
        decay (str): can be either 'lin' or 'exp', corresponding to linear or exponential decay.
    Returns:
        array of histograms (numpy.Array with shape (n_cells, n_pols, surface_size, surface_size))
    """

    assert surface_size <= cell_size
    assert surface_size % 2 != 0
    assert "x" and "y" and "t" and "p" in events.dtype.names
    assert decay == "lin" or decay == "exp"

    # Organizing the events in cells which are, then, saved as NumPy arrays.
    locmems = _map_to_locmems(events, sensor_size, cell_size)
    w, h, npols = sensor_size
    ncells = int(ceil(w / cell_size) * ceil(h / cell_size))
    hist = np.zeros((ncells, npols, surface_size, surface_size), dtype=np.float32)
    # Now we have fun: we cycle on the local memories and generate the histogram corresponding to each of them.
    for c in range(ncells):
        for p in range(npols):
            hist[c, p, :, :] = np.sum(
                np.stack(
                    [
                        _get_ts(
                            locmems[c][p][i],
                            locmems[c][p][:i],
                            time_window,
                            tau,
                            surface_size,
                            decay,
                        )
                        for i in range(len(locmems[c][p]))
                    ]
                    if locmems[c][p].size > 0
                    else np.zeros((surface_size, surface_size), dtype=np.float32)
                ),
                axis=0,
            ) / max(1, locmems[c][p].size)
    return hist



from dataclasses import dataclass

import numpy as np
import torchdata.datapipes.iter as pipes
from torchdata.datapipes import functional_datapipe


@dataclass
@functional_datapipe("slice_by_time")
class SliceByTime(pipes.IterDataPipe):
    """Slices an event array along fixed time window and overlap size. The number of bins depends
    on the length of the recording. Only works on numpy event arrays that contain a 't' or 'ts'
    field.

    >        <overlap>
    >|    window1     |
    >        |   window2     |

    Parameters:
        time_window (int): time for window length (same unit as event timestamps)
        overlap (int): overlap (same unit as event timestamps)
        include_incomplete (bool): include the last incomplete slice that has shorter time
    """

    source_dp: pipes.IterDataPipe
    dt: float
    overlap: float = 0.0
    include_incomplete: bool = False

    def __iter__(self):
        it = iter(self.source_dp)
        while True:
            try:
                events = next(it)
                if "t" in events.dtype.names:
                    t = events["t"]
                elif "ts" in events.dtype.names:
                    t = events["ts"]
                stride = self.dt - self.overlap
                assert stride > 0
                rounding_fn = np.ceil if self.include_incomplete else np.floor
                n_slices = int(rounding_fn(((t[-1] - t[0]) - self.dt) / stride) + 1)
                n_slices = max(n_slices, 1)  # for strides larger than recording time

                window_start_times = np.arange(n_slices) * stride + t[0]
                window_end_times = window_start_times + self.dt
                indices_start = np.searchsorted(t, window_start_times)[:n_slices]
                indices_end = np.searchsorted(t, window_end_times)[:n_slices]
                for start, end in zip(indices_start, indices_end):
                    yield events[start:end]
            except StopIteration:
                return


@dataclass
@functional_datapipe("slice_by_event_count")
class SliceByEventCount(pipes.IterDataPipe):
    """Slices data and targets along a fixed number of events and overlap size. The number of bins
    depends on the amount of events in the recording. Only works on numpy event arrays.

    Parameters:
        event_count (int): number of events for each bin
        overlap (int): overlap in number of events
        include_incomplete (bool): include the last incomplete slice that has fewer events
    """

    source_dp: pipes.IterDataPipe
    n: int
    overlap: int = 0
    include_incomplete: bool = False

    def __iter__(self):
        it = iter(self.source_dp)
        while True:
            try:
                events = next(it)
                n_events = len(events)
                event_count = min(self.n, n_events)
                stride = self.n - self.overlap
                if stride <= 0:
                    raise Exception(
                        "Inferred stride <= 0. Increase n or decrease overlap."
                    )
                rounding_fn = np.ceil if self.include_incomplete else np.floor
                n_slices = int(rounding_fn((n_events - event_count) / stride) + 1)
                for start in (np.arange(n_slices) * stride).astype(int):
                    yield events[start : start + event_count]
            except StopIteration:
                return



import os
from pathlib import Path
from typing import Callable, Iterator, Optional

from expelliarmus import Wizard
from expelliarmus.wizard.clib import event_t
from torchdata.datapipes.iter import FileLister, IterDataPipe

from .utils._dataset import Dataset, Sample
from .utils._utils import check_sha256


class NCARSFileReader(IterDataPipe[Sample]):
    def __init__(
        self,
        dp: IterDataPipe[str],
    ) -> None:
        self.dp = dp
        self._wizard = Wizard(encoding="dat")

    def __iter__(self) -> Iterator[Sample]:
        for fname in self.dp:
            yield (
                self._wizard.read(fname),
                self._get_target(fname),
            )

    def _get_target(self, fname: str) -> int:
        folder_name = fname.split(os.sep)[-2]
        assert (
            folder_name == "background" or folder_name == "cars"
        ), f'Error, the folder name "{folder_name}" is wrong and cannot be associated to a label.'
        return 0 if folder_name == "background" else 1


class NCARS(Dataset):
    """`N-CARS <https://www.prophesee.ai/2018/03/13/dataset-n-cars/>`_

    This datasets needs 'expelliarmus' installed on the system. Events have "txyp" ordering.
    ::

        @article{Sironi_2018_CVPR,
          author = {Sironi, Amos and Brambilla, Manuele and Bourdis, Nicolas and Lagorce, Xavier and Benosman, Ryad},
          title = {HATS: Histograms of Averaged Time Surfaces for Robust Event-Based Object Classification},
          booktitle = {Proceedings of the IEEE Conference on Computer Vision and Pattern Recognition (CVPR)},
          month = {June},
          year = {2018}
        }

    Parameters:
        root (string): Location to decompressed archive.
        train (bool): If True, uses training subset, otherwise testing subset.
    """

    _DTYPE = event_t
    _TRAIN_PATH = "n-cars_train"
    _TEST_PATH = "n-cars_test"
    sensor_size = dict(x=120, y=100, p=2)

    def __init__(
        self,
        root: os.PathLike,
        train: Optional[bool] = True,
        skip_sha256_check: Optional[bool] = True,
    ) -> None:
        self.train = train
        super().__init__(
            Path(root, self.__class__.__name__),
            False,
            skip_sha256_check,
        )
        assert (
            self._check_exists()
        ), "Error: the dataset files could not be found. You should download the dataset and manually extract it and, then, provide the path to the extracted archive as root."

    def __len__(self) -> int:
        return 7482 + 7940 if self.train else 4396 + 4211

    def _filter(self, fname: str) -> bool:
        return fname.endswith(".dat")

    def _check_exists(self) -> bool:
        # Checking that train and test folders exist.
        ret = Path(self._root, self._TRAIN_PATH).is_dir()
        ret = ret and Path(self._root, self._TEST_PATH).is_dir()
        # Checking that some binary files are present.
        cnt = 0
        if ret:
            dp = FileLister(self._root, recursive=True).filter(self._filter)
            for f in dp:
                cnt += 1
        return ret and cnt > 0

    def _datapipe(self) -> IterDataPipe[Sample]:
        fpath = self._TRAIN_PATH if self.train else self._TEST_PATH
        fpath = os.path.join(self._root, fpath)
        dp = FileLister(str(fpath), recursive=True).filter(self._filter)
        # Reading data to structured NumPy array and integer target.
        dp = NCARSFileReader(dp)
        return dp



from pbr.version import VersionInfo

from . import collation, datasets, io, slicers, transforms, utils
from .cached_dataset import CachedDataset, DiskCachedDataset, MemoryCachedDataset
from .dataset import Dataset
from .sliced_dataset import SlicedDataset

try:
    from . import prototype
except (ImportError, NameError):
    pass

all = "__version__"
__version__ = VersionInfo("tonic").release_string()
